#dbutils.py
#$Id: dbutils.py 14745 2004-09-23 17:38:39Z rgbecker $

"""Centralizes all database utilities our apps might need.

This module should provide any high-level conveniences
we need when working with databases.  The kind of
things one would put here include:

The main thing here is a simple object-relational mapping.
This is heavily inspired from Ian Bicking's SQLObjects
(sqlobject.sourceforge.net), and tries to copy the
interface of SQLObject whenever there i a good reason not
to.  However, we did not like some aspects of that package
and did not want the burden of an external dependency.

For examples of use see test_dbutils.py. The idea is
that you create a class inheriting from Table, and
declare fields in it:

class TestPerson(Table):
    _fieldDefs = [
        IntColumn('id',primaryKey=1,allowNull=0),
        TextColumn('firstName',allowNull=0,size=50),
        TextColumn('lastName',allowNull=0,size=50),
        DateTimeColumn('dateOfBirth',allowNull=1)
        ]

Having done this, you can call class methods to
create the tables (if needed) and fetch:
  klass.createTables(conn, dropIfExists=1)

 obj = klass.fetchOne(conn, id)
    This fetches one, identified by primary key.
    
 objectList = klass.fetchMany(conn, **kw)
    This returns a list of objects, ordered by id.
    here keywords can be used as where clauses e.g.
    openBugs = Bug.fetchMany(status='open')
 
 obj.store():
    saves object in database.  If no records exist
    with same primary key, inserts it, else updates it.
    If primary key is called 'id' and is None, generates
    it.

 obj.remove():
    removes object from database



Currently the objects do type conversion and
checking.  Thus,
    >>> fred.id = '7'
    >>> fred.id
    7
This means you can easily process web forms,
CSV files and other sources where 'everything is
a string'.


Limitations:
=============
Right now there is a very simple model where everything
has a single primary key which is integer.  This
should be relaxed to do composite primary keys.

The column definition should allow different column
names, or naming-conversion schemes, to make it
easy to map existing databases.

We need to declare foreign keys and optional table
parameters (e.g. type=InnoDB) in the column definition
so we can create better schemas.

I suspect this fails on drivers other than mx.ODBC.

Some things are being done at instance level which
a metaclass could do better - e.g. pre-generate all
the SQL, let the columns be attributes, generate
optimal methods to read/write delimited files.

Intended Use:
=============
You can subclass Table and add methods and attributes
for your own use freely.  Each object has a connection.
So, you could implement a getSpeakers to do

  talk = Talk.fetch(127)
  for speaker in talk.getSpeakers():
      blah blah

(one day this will be done automatically from
a foreign key traversal)

Also, complex objects may wish to override store().

"""
from pprint import pprint as pp
from types import IntType, StringType, ListType, FloatType, TupleType
from mx.DateTime import DateTimeType, DateTimeFrom



def sqlRepr(thing):
    "Represent how SQL likes it"
    if thing is None:
        return "NULL"
    elif type(thing) is StringType:
        return "'" + thing.replace("'","''") + "'"
    else:
        return str(thing)
    
class Column:
    """Descriptor for a column in a database"""
    def __init__(self, name, allowNull=1, default=None, primaryKey=0):
        self.name = name
        self.allowNull = allowNull
        self.default = default
        self.primaryKey = primaryKey
        
    def getCreateClause(self):
        "get a clause to go in the Create Table statement"
        parts = []
        parts.append('`%s`' % self.name)
        parts.append(self.getSqlTypeDef())
##        if self.primaryKey:   #NO, do this at end
##            parts.append('PRIMARY KEY')
        if self.allowNull:
            parts.append('NULL')
        else:
            parts.append('NOT NULL')
        if self.default is not None:
            parts.append('DEFAULT ' + self.toString(self.default))
        return ' '.join(parts)

    def getSqlTypeDef(self):
        "return the SQL type e.g. 'VARCHAR(50)'"
        return "VARCHAR(50)"

    def toString(self, value):
        if value is None:
            return 'NULL'
        else:
            return str(value)
    
    def assign(self, value):
        "check the value, converting if needed, or raise exception"
        if value is None:
            if not self.allowNull:
                raise ValueError('Column %s cannot accept null values' % self.name)
            else:
                return None
        else:
            try:
                return self.assignNotNull(value)
            except:
                raise ValueError("Unable to assign value %s to field '%s' of type %s" % (value, self.name, self.getSqlTypeDef()))

    def assignNotNull(self, value):
        "Convert value (which is not None) to appropriate type"
        return value
            
class IntColumn(Column):
    def assignNotNull(self, value):
        return int(float(value))

    def getSqlTypeDef(self):
        return "INTEGER"

class FloatColumn(Column):
    def assignNotNull(self, value):
        return float(value)

    def getSqlTypeDef(self):
        return "FLOAT"

        
class BoolColumn(Column):
    def assignNotNull(self, value):
        return int(bool(value))

    def getSqlTypeDef(self):
        return "BOOL"

    def toString(self, value):
        if value is None:
            return 'NULL'
        else:
            if value:
                return '1'
            else:
                return '0'
            

class BlobColumn(Column):

    def assignNotNull(self, value):
        return str(value)

    def getSqlTypeDef(self):
        return "LONGBLOB"

    def toString(self, value):
        if value is None:
            return 'NULL'
        else:
            # encode as hex to avoid having to escape anything
            return "x'%s'" % str(value).encode('hex')


class TextColumn(Column):
    def __init__(self, name, allowNull=1, default=None, primaryKey=0, size=None):
        Column.__init__(self, name, allowNull, default, primaryKey)
        self.size = size

    def assignNotNull(self, value):
        return str(value)[0:self.size]

    def getSqlTypeDef(self):
        if self.size is None:
            return "TEXT"
        else:
            return "VARCHAR(%d)" % self.size

    def toString(self, value):
        if value is None:
            return 'NULL'
        else:
            
            return "'" + str(value).replace("'", "''") + "'"

    
class DateTimeColumn(Column):
    
    def assignNotNull(self, value):
        if type(value) is DateTimeType:
            return value
        elif type(value) is StringType:
            return DateTimeFrom(value)
            

    def getSqlTypeDef(self):
        return "DATETIME"

    def toString(self, value):
        if value is None:
            return 'NULL'
        else:
            return "'" + ' '.join([value.date, value.time]) + "'"


class DateColumn(DateTimeColumn):
    
    def getSqlTypeDef(self):
        return "DATE"

    def toString(self, value):
        if value is None:
            return 'NULL'
        else:
            return "'%s'" % value.date


class TimeColumn(DateTimeColumn):
    
    def getSqlTypeDef(self):
        return "TIME"

    def toString(self, value):
        if value is None:
            return 'NULL'
        else:
            return "'%s'" % value.time


class Index:
    def __init__(self, fieldOrFields, indName=None):
        t = type(fieldOrFields)
        if t in (ListType, TupleType):
            self.fields = fieldOrFields
        elif t is StringType:
            self.fields = [fieldOrFields]
        else:
            raise ValueError, "field parameter to index constructor must be a string field name, or a list/tuple of strings"
        self._indName = indName #allow on the fly creation

    def getName(self):
        if self._indName:
            return self._indName
        else:
            return 'ind_' + '_'.join(self.fields)

    def getCreateClause(self):
        sql = "INDEX `%s` (%s)" % (
            self.getName(),
            ", ".join(['`%s`' % fld for fld in self.fields])
            )
        return sql
        
            

class ForeignKey:
    def __init__(self, thisTableField, refTableName, refTableField):
        self.thisTableField = thisTableField
        self.refTableName = refTableName
        self.refTableField = refTableField

    def getThisTableFields(self):
        if type(self.thisTableField) in (ListType, TupleType):
            return list(self.thisTableField)
        else:
            return [self.thisTableField]

    def getCreateClause(self):
        if type(self.thisTableField) in (ListType, TupleType):
            strThis = '`, `'.join(self.thisTableField)
        else:
            strThis = self.thisTableField

        if type(self.refTableField) in (ListType, TupleType):
            strThat = '`, `'.join(self.refTableField)
        else:
            strThat = self.refTableField
        sql = "FOREIGN KEY (`%s`) REFERENCES `%s`(`%s`)" % (
            strThis,
            self.refTableName,
            strThat
            )
        return sql
    
class ColumnsMetaClass(type):
    """
    Metaclass to provide classes with ready-made defaults and setters.

    Classes which use this metaclass must specify a _fieldDefs class attribute
    which is a list of Column objects.
    For each column 'name' it creates
    (a) an internal variable _name
    (b) a setter function using the column's 'assign' function,
        which can convert from strings and other types
    (c) a getter which returns it trivially
    (d) a public property using the above
    """
    
    def __init__(cls, name, bases, dict):
        super(ColumnsMetaClass, cls).__init__(name, bases, dict)
        for fieldDef in dict['_fieldDefs']:
            pubFieldName = fieldDef.name
            privFieldName = '_' + fieldDef.name

            # these anonymous functions are tricky to define; you have
            # to pass in the privFieldName and fieldDef as keywords,
            # otherwise each such function will get the type of the LAST
            # fieldDef iterated over in the loop.  Must be a better way...
            def setter(self, value, privFieldName = privFieldName, fieldDef=fieldDef):
                self.__dict__[privFieldName] = fieldDef.assign(value)

            def getter(self, privFieldName = privFieldName):
                return self.__dict__[privFieldName]
            
            setattr(cls, pubFieldName, property(getter, setter))    



class Table(object):
    __metaclass__ = ColumnsMetaClass

    # SQL will use class name unless this is set
    _tableName = None

    _tableType = None   #storage type qualifier e.g. 'InnoDb' for MySQL

    # subclasses should set this to list of Column objects    
    _fieldDefs = []
    _foreignKeys = []
    _indices = []

    def getTableName(cls):
        if cls._tableName is None:
            return cls.__name__
        else:
            return cls._tableName
    getTableName = classmethod(getTableName)
    
    def getFieldNames(cls):
        "Return list of field names"
        return [field.name for field in cls._fieldDefs]
    getFieldNames = classmethod(getFieldNames)

    def getPrimaryKey(cls):
        "Return list of fields used in primary key"
        keyFields = []
        for fld in cls._fieldDefs:
            if fld.primaryKey == 1:
                keyFields.append(fld.name)
        if not keyFields:
            raise KeyError("No primary key found!")
        else:
            return keyFields
    getPrimaryKey = classmethod(getPrimaryKey)

    def getPrimaryKeyFieldNames(cls):
        "Return list of fields used in primary key"
        keyFields = []
        for fld in cls._fieldDefs:
            if fld.primaryKey == 1:
                keyFields.append(fld.name)
        assert keyFields, "No primary key found!"
        return keyFields
    getPrimaryKeyFieldNames = classmethod(getPrimaryKeyFieldNames)

    def getPrimaryKeyValues(self):
        "Return list with values in key"
        key = []
        for fld in self._fieldDefs:
            if fld.primaryKey == 1:
                key.append(getattr(self, fld.name))
        assert key, "No primary key found!"
        return key

    def getPrimaryKeyWhereClause(self):
        "Return 2-tuple with fields and values in key"
        where = ' AND '.join(["%s = ?" % name for name in self.getPrimaryKeyFieldNames()])
        params = self.getPrimaryKeyValues()
        return where, params

    def existsInDatabase(self):
        "Does a record with my key(s) exist in database?"
        where, params = self.getPrimaryKeyWhereClause()

        stmt = "SELECT Count(*) FROM %s WHERE %s" % (
            self.getTableName(),
            where)
        cur = self.getConnection().cursor()
        cur.execute(stmt, tuple(params))
        count = cur.fetchone()[0]
        return count == 1
    
    def fetchOne(cls, conn, id):
        "Return one object, with given primary key"
        if type(id) not in (ListType, TupleType):
            params = [id]
        else:
            params = id
            
        fieldNames = map(lambda fld: fld.name, cls._fieldDefs)
        whereClause = ' AND '.join(["%s = ?" % name for name in cls.getPrimaryKeyFieldNames()])
        sql = "SELECT %s FROM %s WHERE %s" % (
            ", ".join(fieldNames),
            cls.getTableName(),
            whereClause
            )
        cur = conn.cursor()
        cur.execute(sql, tuple(params))
        data = cur.fetchall()
        
        assert len(data) > 0, 'No data returned'
        row = data[0]
        obj = cls()
        obj._conn = conn
        for (fieldDef, value) in zip(cls._fieldDefs, row):
            #go straight into dict, we assume database
            #is correct so go at full speed
            obj.__dict__['_' + fieldDef.name] = value
        return obj

    fetchOne = classmethod(fetchOne)            
    
    def fetchMany(cls, conn, **kw):
        """Return a list of objects in primary key order.

        The keywords will be applied as where-clauses.
        """        
        whereClauses = []
        params = []

        for key, value in kw.items():
            whereClauses.append("%s = ?" % key)
            params.append(value)
            
        if whereClauses:
            wherePart = 'WHERE %s' % ' AND '.join(whereClauses)
        else:
            wherePart = ''
        sql = """
            SELECT %s
            FROM %s
            %s
            ORDER BY %s
        """ % (
            ','.join([fieldDef.name for fieldDef in cls._fieldDefs]),
            cls.getTableName(),
            wherePart,
            ', '.join(cls.getPrimaryKeyFieldNames())
        )

        return cls.queryMany(conn, sql, tuple(params))

    fetchMany = classmethod(fetchMany)            

    def fetchManyLike(cls, conn, **kw):
        """Return a list of objects in primary key order.

        The keywords will be applied as where-clauses, but
        rather than searching for equality we pattern-match.
        It would be nice to be able to mix them both, but
        perhaps this is a YAGNI. Andy will probably see a
        way to factor this in more sensibly.    11/02/03 SH
        """        
        whereClauses = []
        params = []

        for key, value in kw.items():
            whereClauses.append("%s LIKE ?" % key)
            params.append(value)
            
        if whereClauses:
            wherePart = 'WHERE %s' % ' AND '.join(whereClauses)
        else:
            wherePart = ''
        sql = """
            SELECT %s
            FROM %s
            %s
            ORDER BY %s
        """ % (
            ','.join([fieldDef.name for fieldDef in cls._fieldDefs]),
            cls.getTableName(),
            wherePart,
            ', '.join(cls.getPrimaryKeyFieldNames())
        )

        return cls.queryMany(conn, sql, tuple(params))

    fetchManyLike = classmethod(fetchManyLike)            

    def objectFromRowTuple(cls, conn, row):
        obj = cls()
        obj._conn = conn
        for (fieldDef, value) in zip(cls._fieldDefs, row):
            #go straight into dict, we assume database
            #is correct so go at full speed
            obj.__dict__['_' + fieldDef.name] = value
        return obj

    objectFromRowTuple = classmethod(objectFromRowTuple)

    def queryMany(cls, conn, sql, argstuple=()):
        """Return row objects from arbitrary SQL query."""
        cur = conn.cursor()
        cur.execute(sql, argstuple)
        data = cur.fetchall()
        result = []
        for row in data:
            result.append(cls.objectFromRowTuple(conn, row))
        return result

    queryMany = classmethod(queryMany)

    def insertRows(cls, conn, fieldNames, dataSet):
        """Fast bulk insertion for CSV and similar imports.

        Let's say you have a multi-column CSV file with
        columns that correspond to fields for a Table subclass.
        Read this into a Python 2d array (e.g. list-of-lists)
        with the columns you want.  Also, prepare a simple
        list of which field names they correspond to.
        If you're really lucky this might be row 0 of your
        file.

        This method will convert/check every value, but
        will then call the DB connection 'executemany'
        method, using only one trip across the network
        instead of thousands.

        It presumes you have set the keys or they meet the
        criteria for auto-generation; if a record exists
        already there will be an exception.
        """        

        #look up and cache the convert functions
        convertersByName = {}
        for fld in cls._fieldDefs:
            func = fld.assign
            name = fld.name
            convertersByName[name] = func
        converters = []
        for fieldName in fieldNames:
            converters.append(convertersByName[fieldName])
        

        rowNo = 0
        colCount = len(fieldNames)
        cleanSet = []
        for row in dataSet:
            cleanRow = []
            for colNo in range(colCount):
                value = row[colNo]
                func = converters[colNo]
                cleaned = func(value)
                cleanRow.append(cleaned)
            cleanSet.append(tuple(cleanRow))
        cur = conn.cursor()
        sql = "INSERT INTO %s (%s) VALUES (%s)" % (
            cls.getTableName(),
            ", ".join(fieldNames),
            ", ".join(["?"] * len(fieldNames))
            )
        cur.executemany(sql, cleanSet)
        cur.close()
        conn.commit()

    insertRows = classmethod(insertRows)    
        
    def __init__(self, conn = None, **kw):
        self._conn = conn
        for fld in self._fieldDefs:
            #print 'initialising %s -> %s' % (fld.name, fld.default)
            self.__dict__['_' + fld.name] = fld.default
        if kw:
            for key, value in kw.items():
                setattr(self, key, value)
                
    def setConnection(self, conn):
        self._conn = conn

    def getConnection(self):
        return self._conn
    
    def setFieldValues(self, **values):
        for (name, value) in values.iteritems():
            # currently unknown field names will just set the attribute
            # on the instance as normal - should probably do something
            # else
            setattr(self, name, value)

    def _getNextId(self):
        cur = self._conn.cursor()
        cur.execute('SELECT MAX(id) FROM %s' % self.getTableName())
        maxId = cur.fetchone()[0]
        cur.close()
        if maxId is None:
            return 1
        else:
            return int(maxId) + 1

    def update(self):
        "Update an existing record, return list of primary key fields"
        cur = self._conn.cursor()
        try:
            # got id attribute, update the matching record
            fieldNames = [field.name for field in self._fieldDefs]
            fieldValues = [getattr(self, name) for name in fieldNames]
            whereClause, keyParams = self.getPrimaryKeyWhereClause()
            
            sql = "UPDATE %s SET %s WHERE %s" % (
                self.getTableName(),
                ','.join(['%s=?' % fieldName for fieldName in fieldNames]),
                whereClause
            )
            params = tuple(fieldValues + keyParams)
            cur.execute(sql, params)
        finally:
            cur.close()
        return keyParams

    def insert(self):
        "Insert a new record, return list of primary key fields"
        cur = self._conn.cursor()
        try:
            if self.getPrimaryKeyFieldNames() == ['id'] and self.id is None:
                #autonumber
                self.id = self._getNextId()

            fieldNames = [field.name for field in self._fieldDefs]
            fieldValues = [getattr(self, name) for name in fieldNames]
            sql = "INSERT INTO %s (%s) VALUES (%s)" % (
                self.getTableName(),
                ','.join(fieldNames),
                ','.join(['?'] * len(fieldValues))
            )
            cur.execute(sql, tuple(fieldValues))
            keyParams = self.getPrimaryKeyValues()
        finally:
            cur.close()
        return keyParams

    def store(self):
        "Insert or update, returns primary key."        
        if self.existsInDatabase():
            keyParams = self.update()
        else:
            keyParams = self.insert()

        if len(keyParams) == 1:
            return keyParams[0]
        else:
            return tuple(keyParams)


    def remove(self):
        where, keyParams = self.getPrimaryKeyWhereClause()
        sql = "DELETE FROM %s WHERE %s" % (self.getTableName(), where)
        cur = self._conn.cursor()
        cur.execute(sql, tuple(keyParams))

    
    def dropTable(cls, conn=None):
        "Generate 'drop table' statement, and execute if cursor passed in"
        dropStmt = 'DROP TABLE IF EXISTS `%s`;' % cls.getTableName()
        if conn:
            cur = conn.cursor()
            try:
                result = cur.execute(dropStmt)
            finally:
                cur.close()
        else:
            result = dropStmt
        return result
    dropTable = classmethod(dropTable)


    def createTable(cls,
                    conn=None,
                    dropIfExists=0,
                    ifNotExists=0):
        "Generate 'create table' statement, and execute if cursor passed in"

        parts = []
        if conn:  #cannot do multiple statements in one call
            cur = conn.cursor()
        
        if dropIfExists:
            dropResult = cls.dropTable(conn)
            if not conn:
                parts.append(dropResult)
        parts.append('CREATE TABLE `%s` (' % cls.getTableName())
        for field in cls._fieldDefs:
            parts.append('    ' + field.getCreateClause() + ',')
        #make primary key clause separate
        parts.append(
            '    PRIMARY KEY PrimaryKey (%s),' % ', '.join(
                ['`%s`' % fld for fld in cls.getPrimaryKeyFieldNames()]
            )
        )

        indexedCols = {}
        # If primary key is single field, include it in list of indices
        if len(cls.getPrimaryKeyFieldNames()) == 1:
            indexedCols[cls.getPrimaryKeyFieldNames()[0]] = 1
            
        #create any explicit indices
        if cls._indices:
            for index in cls._indices:
                parts.append('    %s,' % index.getCreateClause())
                for fieldName in index.fields:
                    indexedCols[fieldName] = 1

        if cls._foreignKeys:
            # first, for each column which is in a foreign key,
            # ensure it has an index.  The primary key will
            # do but if it is not there, it needs an explicit index.

            for fKey in cls._foreignKeys:
                for fld in fKey.getThisTableFields():
                    #if not indexed, make one
                    if not indexedCols.has_key(fld):
                        #add one
                        index = Index(fld)
                        #cls._indices.append(index)
                        parts.append('    %s,' % index.getCreateClause())
                        indexedCols[fld] = 1

                parts.append('    ' + fKey.getCreateClause() + ',')
                
        if parts[-1][-1] == ',':
            parts[-1] = parts[-1][0:-1]  #chop off the last comma
        parts.append('    )')
        if cls._tableType:
            parts[-1] = parts[-1] + ' TYPE=%s' % cls._tableType
            
        parts[-1] = parts[-1] + ';'
        sql = '\n'.join(parts)
        if conn is None:
            return sql
        else:
            try:
                retval = cur.execute(sql)
            except:
                print 'sql trace before error...'
                print sql
                raise
            cur.close()
            return retval
    createTable = classmethod(createTable)
            

