# security handler
"""
This provides a generic login manager facility.

Many apps exhibit a fairly basic, standard security model.
You need to log in to some application; are issued some
token which is good for a period of time; and will/may be
logged out after a period of inactivity.  This is commonly
implemented by a database, or by a directory of temporary
files.

This module provides a base SecurityHandler class, and
some concrete implementations using file system and
databases. You can also subclass it to hook into some
other in-house authentication system.

At run time in a web app, you will typically call 'login'
to get back a Session object.  This has basic info about
when the session started, and a User object.

>>> mgr = loginmanager.getTestManager()
>>> session = mgr.login('guido','possum')
>>> session.token
'46a830b34c80d2394364b5cf34f8940b'
>>> session.started, session.touched
(1025563900.938, 1025563900.938)
>>> session.user
<rlextra.ers.loginmanager.User instance at 013238A4>
>>> session.user.userName
'guido'
>>> session.user.hashPass
'30983b5377baa2633477cb4743ca7168'

User and Session are pretty minimal; the intent is that
you can hang extra attributes of them if you wish. At
the end of a CGI, or at any point, you can call 
   mgr.saveSession(session)
This will update the timestamp and also persist the
session object by whatever means (e.g. file system pickle).
Take great care not to have your entire object model
hanging off the user or session object or you'll create
a monster.

Note that the User object stores the hash password and
not the real one, and the real password is not held in
memory longer than needed.  You can also log in with
this format :
>>> session = mgr.login('guido',
        hashPassword='30983b5377baa2633477cb4743ca7168')

Many things are not done.  In particular there is no
concept of users having access rights to resources; all this
does is see who was logged in.  Once we get to that level,
every app is different; you can easily add new methods
to subclasses.


TO DO:
  - write lots of test cases
  - do a database implementation
  - deploy in anger

"""
import md5, time, socket, random, os, string, cPickle, glob, base64, cStringIO

class SecurityError(Exception):
    pass

class TimeOutError(Exception):
    pass

def _notValidated():
    raise SecurityError("Authorisation not validated!")

class Session:
    "A general 'data packet' you can attach anything to"
    def __init__(self, user):
        self.user = user
        self.started = time.time()
        self.touched = time.time()

    def touch(self):
        "Update my 'last used' time"
        self.touched = time.time()

def dict2base64(d):
    '''convert a dict into base64'''
    return base64.encodestring(cPickle.dumps(d,1))[:-1]

def base642dict(b):
    '''convert a base64 string into a dict'''
    return cPickle.loads(base64.decodestring(b))

class User:
    "Holds a user name, password or whatever."
    def __init__(self, userName):
        self.userName = userName
        self.hashPass = None
        self.realName = None

class SecurityHandler:
    def __init__(self, timeOut=None):
        self.timeOut = timeOut
        
    def generateToken(self):
        """Generates a 'reasonably unique' string token"""
        # make some string up highly unlikely to be generated
        # by another system.  I think any 2 of these three
        # would probably do.
        plain = '%s-%s-%s' % (socket.gethostname(),
                              time.time(),
                              random.random())
        # return something containing nice safe characters
        # which URLs, browsers etc. won't scramble:  0-9 and A-F!
        scrambled = md5.md5(plain).hexdigest()
        return scrambled
    
    def addUser(self, userName, password, realName=None):
        "Use this if you want to; not normally called at run time"
        pass

    def changePassword(self, userName, newPassword):
        pass

    def revokeUser(self, user):
        "Delete or ban them, whatever"
        pass
    
    def login(self, userName, password='',hashPassword=None):
        """Log in using either a plain or hashed password

        Either returns a token, or raises a SecurityError.
        """
        # this one always succeeds
        return self.generateToken()

    def logout(self, token):
        pass

    def purge(self):
        "Housekeeping hook to purge expired sessions"
        
    def getSession(self, token):
        """Check if a token is valid and retrieve a session
        object.  If timed out or token is not valid, raise
        an exception"""
        return Session(User('bogus',''))

    def saveSession(self, session):
        """Override this to save the session somehow.
        Can save any Python object"""
        pass

class MemorySecurityHandler(SecurityHandler):
    """Providers a standard authentication scheme,
    but forgets everything when deleted. Useful for
    testing and exploring the concept and as a basis for
    file and database security handlers."""
    def __init__(self, timeOut=None):
        self._users = {}
        self._sessions = {}
        SecurityHandler.__init__(self, timeOut=timeOut)

    def addUser(self, userName, password='', hashPassword=None, realName=None, role=None):
        "allows you to add a user - supplying either a password or a hashed password"
        u = User(userName)
        u.realName = realName
        if password !='' and hashPassword != None:
            raise "both plain and encrypted password supplied - please use only one."
        if hashPassword !=None:
            encryptedPassword = hashPassword
        elif password !='':
            encryptedPassword = md5.md5(password).hexdigest()
        u.hashPass = encryptedPassword
        u.role = role
        self._users[userName] = u

    def changePassword(self, userName, newPassword):
        u = self._users[userName]
        u.hashPass = md5.md5(password).hexdigest()

    def revokeUser(self, userName):
        if self._users.has_key(userName):
            del self._users[userName]

    def login(self, userName, password='',hashPassword=None):
        # for this implementation we just ask if they
        # are a known user.  Other implementations would
        # probably read some file, if they had not done
        # so on initialisation.
        if not userName or (not password and not hashPassword): _notvalidated()
        try:
            userObj = self._users[userName]
        except KeyError:
            _notValidated()
        if hashPassword is None: hashPassword = md5.md5(password).hexdigest()
        if userObj.hashPass!=hashPassword: _notValidated()

        token = self.generateToken()
        sesh = Session(userObj)
        sesh.token = token
        self._sessions[token] = sesh
        return sesh

    def logout(self, token):
        del self._sessions[token]

    def purge(self):
        "Housekeeping hook to purge expired sessions"
        if self.timeOut is not None:
            now = time.time()
            for k, sesh in self._sessions.items():
                if now - sesh.touched >= self.timeOut:
                    del self._sessions[k]

    def checkTimeOut(self, session):
        "Has it expired?  Raise TimeOutException or return None"
        if self.timeOut is not None:
            #check if timed out
            howLongAgo = time.time() - session.touched
            if howLongAgo > self.timeOut:
                raise TimeOutError("This session for '%s' has expired "
                                   "%d seconds ago; timeout is %d seconds" % (
                                       session.user.userName,
                                       howLongAgo - self.timeOut,
                                       self.timeOut
                                       )
                                   )

    def getSession(self, token):
        "This is typically called on each hit"
        try:
            sesh = self._sessions[token]
        except KeyError:
            raise SecurityError("Token '%s' not valid!" % token)
        self.checkTimeOut(sesh)
        sesh.touch()
        return sesh

    def saveSession(self, session):
        "No work to do here as we aren't persisting anything."
        session.touch()
        
class FileSecurityHandler(MemorySecurityHandler):
    """Security handler which works by saving files in a directory.

    This creates a file called 'users' containing one line per
    user, in the format username:hashPass:'real name'; and
    one file per session whose name is the token and whose
    contents are the pickled session object.  
    """

    def __init__(self, directory=None, timeOut=None):
        if directory is None:
            directory = os.getcwd()
        self.directory = directory
        MemorySecurityHandler.__init__(self, timeOut=timeOut)
        self._loadUsers()

    def _loadUsers(self):
        """Slurp in the lot when initialized.  This is a greedy
        strategy and for CGI and many users we'd want to do better
        and just load on demand"""
        userFileName = self.directory + os.sep + 'users.txt'
        if not os.path.isfile(userFileName):
            #create it; might as well find out as early as
            #possible if the directory is unwriteable!
            open(userFileName, 'w').write('#users file created at ' + time.ctime(time.time()))
        lines = open(userFileName).readlines()
        for line in lines:
            stripped = string.strip(line)
            if len(stripped) > 0 and stripped[0] <> '#':
                (userName, hashPass, realName) = string.split(stripped, ':')
                u = User(userName)
                u.hashPass = hashPass
                u.realName = realName
                self._users[userName] = u

    def _saveUsers(self):
        """Just rewrite the file - again we could be more efficient
        for thousands of them"""
        userFileName = self.directory + os.sep + 'users.txt'
        items = self._users.items()
        items.sort()
        lines = []
        lines.append('#users file modified at ' + time.ctime(time.time()))
        for (key, u) in items:
            lines.append('%s:%s:%s' % (u.userName, u.hashPass, u.realName))
        text = string.join(lines, '\n')
        open(userFileName, 'w').write(text)
        
    def addUser(self, userName, password='', hashPassword=None, realName=None, role=None):
        if password !='' and hashPassword != None:
            raise "both plain and encrypted password supplied - please use only one."
        if hashPassword !=None:
            encryptedPassword = hashPassword
        elif password !='':
            encryptedPassword = md5.md5(password).hexdigest()
        MemorySecurityHandler.addUser(self, userName, hashPassword=encryptedPassword, realName=realName, role=role)
        self._saveUsers()

    def getSession(self, token):
        filename = self.directory + os.sep + token + '.tok'
        if not os.path.isfile(filename):
            raise SecurityError("Token '%s' not valid!" % token)
        else:
            sesh = cPickle.load(open(filename, 'rb'))
            assert sesh.token == token, 'bad session token on disk!'
            self.checkTimeOut(sesh)
            sesh.touch()
            return sesh

    def saveSession(self, session):
        filename = self.directory + os.sep + session.token + '.tok'
        f = open(filename, 'wb')
        session.touch()
        cPickle.dump(session, f)
        f.close()

    def login(self, userName, password='',hashPassword=None):
        #need to save a session to disk when created
        sesh = MemorySecurityHandler.login(self, userName, password, hashPassword)
        self.saveSession(sesh)
        return sesh

    def logout(self, token):
        filename = self.directory + os.sep + token + '.tok'
        os.remove(filename)
        if self._sessions.has_key(token):
            del self._sessions[token]
        
    def purge(self):
        "Housekeeping hook to purge expired sessions"
        if self.timeOut is not None:
            filenames = glob.glob(self.directory + os.sep + '*.tok')
            now = time.time()
            for filename in filenames:
                sesh = cPickle.load(open(filename,'rb'))
                if now - sesh.touched >= self.timeOut:
                    os.remove(filename)
                    if self._sessions.has_key(sesh.token):
                        del self._sessions[sesh.token]


class ODBCSecurityHandler(MemorySecurityHandler):
    """Security handler which works with an mx.ODBC-like database. 

    Based mainly on FileSecurityHandler (above).    

    When initialized, it requires a valid database connection to be passed to it.
    Most methods also required that a valid database is passed to them.
    This works with mx.ODBC connection objects.  Other database connection
    libraries use different sequences, so the actual SQL calls are all broken
    out into methods beginning with '_sql', to allow subclassing.  Yuk, but
    necessary.

    This is the SQL used to create the tables required for DbSecurityHandler to work:

    CREATE TABLE WebUsers (
        UserId varchar (20) PRIMARY KEY NOT NULL ,
        FullName varchar (50) NULL ,
        Role varchar (20) NOT NULL ,
        EncryptedPassword varchar (50) NULL ,
        LastLogin datetime NULL ,
        Enabled bit NOT NULL
    )
    ;

    CREATE TABLE WebSessions (
        UserId varchar (20) NOT NULL ,
        EncryptedPassword varchar (50) NOT NULL ,
        SessionToken varchar (50) PRIMARY KEY NOT NULL ,
        SessionObject text NOT NULL,
    )
    ;

    """

    def __init__(self, db, userTable='Users', sessionTable='Sessions',
                 directory=None, timeOut=None):
        if directory is None:
            directory = os.getcwd()
        self.directory = directory # left in in case we need logging etc
        self._db = db
        self._cur = None
        self._userTable = userTable         # override these
        self._sessionTable = sessionTable   # if your database has different table names

        MemorySecurityHandler.__init__(self, timeOut=timeOut)

    def _getCursor(self):
        "Returns an open database cursor object"
        if self._cur is None:
            self._cur = self._db.cursor()
        return self._cur

    def _commit(self):
        "commits to the database"
        self._db.commit()


    def _sqlInsertUser(self, userName, realName, role, encryptedPassword, enabled):
        "statement for addUser below, subclass this if needed"
        self._cur.execute("INSERT INTO %s (UserId, FullName, Role, EncryptedPassword, Enabled) VALUES (?,?,?,?,?)" % self._userTable,
                    (userName, realName, role, encryptedPassword, 1)
                    )
        
    def addUser(self, userName, password='', hashPassword=None, realName=None, role=None):
        "Add a user by writing to a database"
        cur = self._getCursor()
        if password !='' and hashPassword != None:
            raise "both plain and encrypted password supplied - please use only one."
        if hashPassword !=None:
            encryptedPassword = hashPassword
        elif password !='':
            encryptedPassword = md5.md5(password).hexdigest()
        else:
            raise SecurityError("No valid password supplied for %s" % userName)
        self._sqlInsertUser(userName, realName, role, encryptedPassword, 1)
        self._commit()


    def _sqlSelectUserRole(self, userName):
        self._cur.execute("SELECT userId, Role FROM %s WHERE UserId=?" % self._userTable, (userName,))

    def _sqlUpdatePassword(self, hashPass, userName):
        self._cur.execute("UPDATE %s SET EncryptedPassword=? WHERE UserId=?"%self._userTable,(hashPass, userName))

    def changePassword(self, userName, newPassword):
        cur = self._getCursor()
        self._sqlSelectUserRole(userName)
        retrieved = cur.fetchall()
        if len(retrieved)<1:
            raise SecurityError("user '%s' not found or not allowed!" % userName)
        hashPass = md5.md5(newPassword).hexdigest()
        self._sqlUpdatePassword(hashPass, userName)
        self._commit()

    def _sqlGetSessionToken(self, token):
        self._cur.execute("SELECT SessionToken,SessionObject from %s WHERE SessionToken=?"%self._sessionTable,(token,))
        
    def getSession(self, token):
        """Check if a token is valid and retrieve a session object
        from the database. If timed out or token is not valid, raise
        an exception"""
        cur = self._getCursor()
        self._sqlGetSessionToken(token)
        retrieved = cur.fetchall()
        if len(retrieved)<1:
            raise SecurityError("Token '%s' not valid!" % token)
        rToken, rSesh = retrieved[0]
        sesh = cPickle.loads(rSesh)
        assert sesh.token == token, 'bad session token in database!'
        self.checkTimeOut(sesh)
        self.updateSession(sesh)
        return sesh

    def _sqlUpdateSession(self, pickledSession, token):
        self._cur.execute("UPDATE %s SET SessionObject=? WHERE SessionToken=?" % self._sessionTable,
                                (pickledSession, token))
    def updateSession(self, session):
        session.touch()
        self._sqlUpdateSession(cPickle.dumps(session),session.token)
        self._commit()

    def _sqlSaveSession(self, token, pickledSession):
        self._cur.execute("INSERT INTO %s (SessionToken, SessionObject) VALUES (?,?)" % self._sessionTable,
                    (token, pickledSession)
                    )

    def saveSession(self, session):
        session.touch()
        self._sqlSaveSession(session.token, cPickle.dumps(session))
        self._commit()

    def _sqlGetUserInfo(self, userName):
        self._cur.execute("SELECT userId, FullName, encryptedPassword, Enabled FROM %s WHERE UserId=?" %self._userTable, (userName,))

    def _sqlSetLastLogin(self, timeStamp, userName):        
        self._cur.execute("UPDATE %s SET LastLogin=? WHERE UserId=?"%self._userTable,(timeStamp,userName))

    def login(self, userName, password='',hashPassword=None):
        """Connects to the database, checks that both username and
        password are valid, then returns a session object"""
        if not userName or (not password and not hashPassword): _notValidated()
        cur = self._getCursor()
        self._sqlGetUserInfo(userName)
        retrieved = cur.fetchall()
        if len(retrieved)<1: retrievedUserName = retrievedFullName = retrievedEncryptedPassword = enabled = None
        else: retrievedUserName, retrievedFullName, retrievedEncryptedPassword, enabled = retrieved[0]
        if not enabled or retrievedUserName!=userName: _notValidated()
        if hashPassword is None: hashPassword = md5.md5(password).hexdigest()
        if retrievedEncryptedPassword != hashPassword: _notValidated()
        userObj = User(retrievedUserName)
        userObj.hashPass = hashPassword
        userObj.realName = retrievedFullName
        self._sqlSetLastLogin(self.getTimeStamp(), userName)
        token = self.generateToken()
        #print "login: token: ", token
        sesh = Session(userObj)
        #print "login: sesh: ", sesh
        sesh.token = token
        self._sessions[token] = sesh
        self.saveSession(sesh)
        return sesh

    def _delSession(self,token):
        cur = self._getCursor()
        cur.execute("DELETE FROM %s WHERE SessionToken=?"%self._sessionTable, (token,))
        self._commit()

    def logout(self, token):
        """removes the session token (and the rest of the row it is on)
        from the database sessions table"""
        #database bit...
        self._delSession(token)
        #remove from local dictionary...
        if self._sessions.has_key(token): del self._sessions[token]
        
    def purge(self):
        """A housekeeping function - removes expired sessions from the database
        (and from the local dictionary as well)."""

        if self.timeOut is not None:
            now = time.time()
            cur = self._getCursor()
            cur.execute("SELECT * FROM %s" % self._sessionTable)
            retrieved = cur.fetchall()
            for record in retrieved:
                retrievedName, retrievedEncryptedPassword, retrievedToken, retrievedSession = record
                sesh = cPickle.loads(retrievedSession)
                if now - sesh.touched >= self.timeOut:
                    cur = self._getCursor()
                    cur.execute("DELETE FROM %s WHERE SessionToken=%s" % self._sessionTable, (retrievedToken,))
                    #print "removed token '%s' for user '%s'" % (retrievedToken,retrievedName)
                    self._commit()
                    if self._sessions.has_key(sesh.token):
                        del self._sessions[sesh.token]

    def getTimeStamp(self):
        "nearest second - our databases cannot handle microseconds!"
        from mx import DateTime 
        return DateTime.gmtime(int(time.time()))


class MySQLSecurityHandler(ODBCSecurityHandler):
    """Security handler using MySQLdb API

    This is just different enough to be annoying, and to
    require that we override all the SQL.  %s collides with
    our need to do string substitutions on the table name,
    so we tend to buid explicit SQL statemements."""
    def _sqlInsertUser(self, userName, realName, role, encryptedPassword, enabled):
        "statement for addUser below, subclass this if needed"
        self._cur.execute("INSERT INTO %s (UserId, FullName, Role, EncryptedPassword, Enabled) VALUES (%s,%s,%s,%s,%s)" % self._userTable,
                    (userName, realName, role, encryptedPassword, 1)
                    )

    def _sqlSelectUserRole(self, userName):
        self._cur.execute("SELECT userId, Role FROM %s WHERE UserId='%s'" % (self._userTable, userName))

    def _sqlUpdatePassword(self, hashPass, userName):
        self._cur.execute("UPDATE %s SET EncryptedPassword='%s' WHERE UserId='%s'" % (self._userTable,hashPass, userName))

    def _sqlGetSessionToken(self, token):
        self._cur.execute("SELECT SessionToken,SessionObject from %s WHERE SessionToken='%s'" % (self._sessionTable,token))

    def _sqlUpdateSession(self, pickledSession, token):
        self._cur.execute("UPDATE " + self._sessionTable + " SET SessionObject=%s WHERE SessionToken=%s",  (pickledSession, token))

    def _sqlSaveSession(self, token, pickledSession):
        self._cur.execute("INSERT INTO " + self._sessionTable + " (SessionToken, SessionObject) VALUES (%s,%s)",
                    (token, pickledSession)
                    )

    def _sqlGetUserInfo(self, userName):
        self._cur.execute("SELECT userId, FullName, encryptedPassword, Enabled FROM %s WHERE UserId='%s'"  % (self._userTable, userName))

    def _sqlSetLastLogin(self, timeStamp, userName):        
        self._cur.execute("UPDATE %s SET LastLogin=%s WHERE UserId='%s'" % (self._userTable,timeStamp,userName))
