
"""This module provides generic Steps which can be used by web applications.

The Steps can be sequenced together using the Sequence Step and then used
as webApp 'actions' by wrapping with AdaptToAction (see usage in ConfKit).

"""

# Generic actions ##########################################################

class ParamInfo(object):

    def __init__(self, name, desc = None, optional = False):
        self._name = name
        self._desc = desc
        self._optional = optional

    def getSummary(self):
        name = self._name
        if self._optional:
            name = '[%s]' % name
        desc = self._desc
        if not desc:
            desc = ''
        return '%s: %s' % (name, desc)

    name = property(lambda self: self._name)

class Step:

    name = 'Unnamed'    

    def __call__(self, app, request, response):
        pass

    def setInputs(self, inputs):
        self._inputs = inputs

    def getInputs(self):
        return getattr(self, '_inputs', ())

class DoView(Step):

    def __init__(self, viewName):
        import os
        self._viewDir = os.path.dirname(viewName)
        self._viewName = os.path.basename(viewName)
    
    def __call__(self, app, request, response):
        # SHOULDN'T CALL THIS - it's private, will do for testing though
        response.write(app._getPreppyModOutput(self._viewName, self._viewDir, request, response))

class RedirectResponse(Step):

    def __init__(self, requestParam, step):
        self._requestParam = requestParam
        self._step = step

    def __call__(self, app, request, response):
        import StringIO
        tempResponse = StringIO.StringIO()
        try:
            self._step(app, request, tempResponse)
            result = tempResponse.getvalue()
        finally:
            tempResponse.close()
        if self._requestParam:
            request.params[self._requestParam] = result
        else:
            return result

class RedirectResponseToFile(Step):

    def __init__(self, filename, step):
        self._filename = filename
        self._step = step

    def __call__(self, app, request, response):
        f = open(self._filename, 'w')
        try:
            self._step(app, request, f)
        finally:
            f.close()

class ApplyTemplate(Step):

    def __init__(self, step, title = None):

        self._step = step        
        self._title = title

    def _doHeader(self, app, request, nav, kw):
        kw.setdefault('title', self._title)
        kw.setdefault('status', request.params.get('status', ''))
        kw.setdefault('nav', nav)
        self._doneHeader = 1
        return app.doHeader(**kw)

    def _doFooter(self, app, request, kw):
        self._doneFooter = 1
        return app.doFooter(**kw)
        
    def __call__(self, app, request, response):
        steps = []
        menuView = request.params.get('menuView', None)
        if menuView:
            nav = RedirectResponse(None, DoView(menuView))(app, request, response)
        request.params['doHeader'] = lambda **kw: self._doHeader(app, request, nav, kw)
        request.params['doFooter'] = lambda **kw: self._doFooter(app, request, kw)
        self._doneHeader = 0
        self._doneFooter = 0
        body = RedirectResponse(None, self._step)(app, request, response)
        if not self._doneHeader:
            response.write(self._doHeader(app, request, nav, {}))
        response.write(body)
        if not self._doneFooter:
            response.write(self._doFooter(app, request, {}))

class RetrieveRecord(Step):

    def __init__(self, table, defaults = None, newRecord = 0, restrictFields = None):
        if restrictFields is None:
            restrictFields = table.getFieldNames()
        self._table = table
        self._defaults = defaults
        self._newRecord = newRecord
        self._restrictFields = restrictFields

    def __call__(self, app, request, response):
        table = self._table
        conn = app.getConnection()
        params = request.params
        defaults = self._defaults

        # Determine if we should fetch an existing record
        newRecord = self._newRecord
        if not newRecord:
            # See if we have all primary key fields...
            keyFields = table.getPrimaryKeyFieldNames()
            keyValues = []
            for keyField in keyFields:
                if not params.has_key(keyField):
                    break
                keyValues.append(params.get(keyField))
            # ... if so, then assume we should fetch existing record
            newRecord = (len(keyValues) != len(keyFields))
            
        if newRecord:
            # return new record
            record = table(conn)
            if defaults:
                for name, value in defaults.iteritems():
                    setattr(record, name, value)
        else:
            # got primary key, fetch record
            record = table.fetchOne(conn, keyValues)

        # Fill in record with any matching parameters                    
        restrictFields = self._restrictFields
        for field in restrictFields:
            if params.has_key(field):
                setattr(record, field, params[field])
                
        request.params['record'] = record        

class EnsureNewRecord(Step):

    """ Check that values for primary key fields are not already present in table """

    def __init__(self, table):
        self._table = table

    def __call__(self, app, request, response):
        table = self._table
        conn = app.getConnection()
        params = request.params

        # Get primary key values
        keyFields = table.getPrimaryKeyFieldNames()
        keyValues = {}
        for keyField in keyFields:
            if not params.has_key(keyField):
                break
            keyValues[keyField] = params.get(keyField)
        
        assert len(keyValues) == len(keyFields), 'Not enough values for primary key %s' % keyValues
            
        # Ensure no mathcing record currently exists
        records = table.fetchMany(conn, **keyValues)
        assert len(records) == 0, 'Record already exists %s' % keyValues

        # Make new object
        record = table(conn)

        # Fill in with any matching parameters                    
        restrictFields = table.getFieldNames()
        for field in restrictFields:
            if params.has_key(field):
                setattr(record, field, params[field])
                
        request.params['record'] = record        

class RetrieveRecords(Step):

    def __init__(self, table):
        self._table = table
        
    def __call__(self, app, request, response):
        records = self._table.fetchMany(app.getConnection())
        request.params['records'] = records        

class StoreRecord(Step):

    def __call__(self, app, request, response):
        record = request.params['record']
        id = record.store()
        app.getConnection().commit()
        
class DeleteRecord(Step):

    def __call__(self, app, request, response):
        record = request.params['record']
        record.remove()
        app.getConnection().commit()

class SetDefaultParams(Step):

    def __init__(self, defaults):
        self._defaults = defaults
        inputs = []
        for param in defaults:
            inputs.append(ParamInfo(param, optional = True))
        self.setInputs(inputs)
        
    def __call__(self, app, request, response):
        for name, value in self._defaults.iteritems():
            request.params.setdefault(name, value)

class CopyParams(Step):

    def __init__(self, params):
        self._params = params
        
    def __call__(self, app, request, response):
        for target, source in self._params.iteritems():
            request.params[target] = request.params[source]

class SetParams(Step):

    def __init__(self, params):
        self._params = params
        
    def __call__(self, app, request, response):
        for name, value in self._params.iteritems():
            request.params[name] = value

class SetParamsFromRecord(Step):

    def __init__(self, params):
        "params can be list of strings or list of sequences"
        self._params = params
        
    def __call__(self, app, request, response):
        import types
        record = request.params['record']
        for paramInfo in self._params:
            if type(paramInfo) in types.StringTypes:
                paramInfo = [paramInfo, paramInfo]
            destParam = paramInfo[0]
            recordParam = paramInfo[1]            
            request.params[destParam] = getattr(record, recordParam)

class Redirect(Step):

    def _isSSL(self):
        """ Return true if we are on a SSL (https) connection. """
        import os
        return os.environ.get('SSL_PROTOCOL', '') != ''    

    # modified from recipe in Python Cookbook...
    def _getScriptURL(self):
        """ Return a full URL starting with schema, servername and port.
            *uri* -- append this server-rooted uri (must start with a slash)
        """
        import os
        schema, stdport = (('http', '80'), ('https', '443'))[self._isSSL()]
        host = os.environ.get('HTTP_HOST', None)
        if not host:
            host = os.environ.get('SERVER_NAME')
            port = os.environ.get('SERVER_PORT', '80')
            if port != stdport: host = host + ":" + port

        return "%s://%s%s" % (schema, host, os.environ.get('SCRIPT_NAME'))

    def __init__(self, forwardParams = None, forwardState = 1):
        self._forwardParams = forwardParams
        if forwardState:
            self._forwardParams.append('token') # CHANGE THIS
        self._scriptURL = self._getScriptURL()
        
    def __call__(self, app, request, response):
        import urllib
        queryParts = []
        for name, value in request.params.iteritems():
            if name in self._forwardParams:
                queryParts.append('%s=%s' % (name, urllib.quote_plus(str(value))))
        newUrl = '%s?%s' % (
            self._scriptURL,
            '&'.join(queryParts)
        )
        response.setHeader('Location', newUrl)

class SetAuthenticationUtils(Step):

    def getHRef(token, **params):
        import urllib
        if token:
            params['token'] = token
        href = '?%s' % urllib.urlencode(params)
        return href
    getHRef = staticmethod(getHRef)

    def __call__(self, app, request, response):
        #TODO: THIS SHOULDN'T HAVE KNOWLEDGE OF AUTHENTICATION DETAILS
        from xml.sax.saxutils import escape
        token = request.params.get('token', None)
        getHRef = lambda token = token, **params : self.getHRef(token, **params) 
        request.params['getRawHRef'] = getHRef
        # Make plain getHRef returns escaped HRef
        request.params['getHRef'] = lambda **kw : escape(getHRef(**kw))

class AddFormAuthentication(Step):

    def __call__(self, app, request, response):
        #TODO: THIS SHOULDN'T HAVE KNOWLEDGE OF AUTHENTICATION DETAILS
        token = request.params.get('token', None)
        if token:
            formAuthentication = '<input type="hidden" name="token" value="%s"/>' % token
            modify = ModifyResponseData(r'(<form\s+.*?>)', r'\1%s' % formAuthentication)
            modify(app, request, response)
    

class ModifyResponseData(Step):

    def __init__(self, regExp, replaceStr):
        import re
        self._regExp = re.compile(regExp)
        self._replaceStr = replaceStr

    def __call__(self, app, request, response):
        # TODO: THIS STUFF RELIES ON RESPONSE INTERNALS
        # MUST CHANGE TO USE SOME (TO BE DEFINED) RESPONSE
        # PUBLIC INTERFACE
        respData = ''.join(response._outputList)
        respData = self._regExp.sub(self._replaceStr, respData)
        response._outputList = [respData]
    
class Sequence(Step):

    def __init__(self, *steps):
        self._steps = steps
        # this is a bit simplistic - currently just assume the inputs to sequence
        # are the same as first step in sequence which specifies any
        NoneFunc = lambda : None
        for step in steps:
            inputs = getattr(step, 'getInputs', NoneFunc)()
            if inputs:
                self.setInputs(inputs)
                break

    def __call__(self, app, request, response):
        for step in self._steps:
            step(app, request, response)  # Sequence

class RecordUpdate(Step):

    def __init__(self, table = None, successAction = None, successStatus = None, restrictFields = None):
        if successStatus is None:
            successStatus = 'Update completed Successfully'
        assert table != None, 'Must supply database table'
        assert successAction != None, 'Must specify successAction'
        idFields = table.getPrimaryKeyFieldNames()
        self._step = Sequence(
            RetrieveRecord(table, restrictFields = restrictFields),
            StoreRecord(),
            SetParamsFromRecord(idFields),
            SetParams({'action': successAction, 'status': successStatus}),
            Redirect(['action', 'status'] + idFields)
        )
        inputs = [ParamInfo('status', 'Status string', optional = True)]
        for fieldName in idFields:
            inputs.append(ParamInfo(fieldName, desc = 'Primary key value for record', optional = True))
        self.setInputs(inputs)

    def __call__(self, app, request, response):
        self._step(app, request, response)
    
class RecordEdit(Step):

    def __init__(self, table = None, recordView = None, title = None, recordDefaults = None):
        if title is None:
            title = 'Untitled'
        if recordDefaults is None:
            recordDefaults = {}
        assert table != None, 'Must supply database table'
        idFields = table.getPrimaryKeyFieldNames()
        steps = [RetrieveRecord(table, recordDefaults)]
        if recordView:
            steps.append(ShowView(recordView, title = title))
        self._step = Sequence(*steps)
        inputs = [ParamInfo('status', 'Status string', optional = True)]
        for fieldName in idFields:
            inputs.append(ParamInfo(fieldName, desc = 'Primary key value for record', optional = True))
        self.setInputs(inputs)

    def __call__(self, app, request, response):
        self._step(app, request, response)

class RecordDelete(Step):

    def __init__(self, table = None, followOnView = None, paramMapping = None):
        assert table != None, 'Must supply database table'
        seq = [RetrieveRecord(table), DeleteRecord()]
        if paramMapping:
            seq.append(SetParamsFromRecord(paramMapping))
        if followOnView:
            seq.append(FollowOnView(followOnView))
        self._step = Sequence(*seq)
        inputs = []
        for fieldName in table.getPrimaryKeyFieldNames():
            inputs.append(ParamInfo(fieldName, desc = 'Primary key value for record'))
        self.setInputs(inputs)

    def __call__(self, app, request, response):
        self._step(app, request, response)

def SetPageTemplate(filename):
    return SetParams({'pageView': filename})
            
class ShowView(Step):

    def __init__(self, viewName, title = None):
        self._viewName = viewName
        self._title = title

    def __call__(self, app, request, response):
        templateFile = request.params.get('pageView')
        Sequence(
            SetParams({'title': self._title}),
            SetDefaultParams({'status': ''}),
            RedirectResponse('body', DoView(self._viewName)),
            DoView(templateFile),
        )(app, request, response)

class AccessDenied(Step):
    def __call__(self, app, request, response):
        response.setHeader('Status','404 Not Found')
        response.write("""
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<html>
   <head>
      <title>Not Found</title>
   </head>
   <body>
      <p>Resource not found.</p>
   </body>
</html>
        """)
    
class SwitchOnRole(Step):
    
    def __init__(self, matchList, noMatch = AccessDenied()):
        self._step = Switch('userRole', matchList, noMatch)

    def __call__(self, app, request, response):
        self._step(app, request, response)
            
class Switch(Step):

    def __init__(self, param, matchList, noMatchStep):
        self._param = param
        self._matchList = matchList
        self._noMatchStep = noMatchStep
        
    def __call__(self, app, request, response):
        nextStep = self._noMatchStep
        if request.params.has_key(self._param):
            switchValue = request.params[self._param]
            for matchValue, matchStep in self._matchList:
                if matchValue == switchValue: 
                    nextStep = matchStep
                    break
        nextStep(app, request, response)
    
class FollowOnView(Step):

    def __init__(self, action):
        self._step = Sequence(
            SetParams({'action': action.name}),
            Redirect(['action'] + [input.name for input in action.getInputs()])
        )
    def __call__(self, app, request, response):
        self._step(app, request, response)

class DecodeParams(Step):
    def __init__(self,paramName):
        self._paramName = paramName
    def __call__(self,app,request,response):
        import base64, marshal#, zlib
        params = request.params
        params.update(marshal.loads(base64.decodestring(params[self._paramName])))
        del params[self._paramName]

class EncodeParams(Step):
    def __init__(self,paramName,paramDefn=None):
        self._paramName = paramName
        self._paramDefn = paramDefn
    def __call__(self,app,request,response):
        params = request.params
        D = {}
        for p in self._paramDefn(app,request,response):
            D[p] = params[p]
        import base64, marshal#, zlib
        params[self._paramName] = base64.encodestring(marshal.dumps(D)).replace('\n','')


class GetFileSize(Step):
    
    def __init__(self, filenameParam, resultParam):
        self._filenameParam = filenameParam
        self._resultParam = resultParam
        
    def __call__(self, app, request, response):
        import os
        params = request.params
        params[self._resultParam] = os.path.getsize(params[self._filenameParam])


class AdaptToAction:

    def __init__(self, app, step, name):
        self._app = app
        self._step = step
        self._step.name = name

    def __call__(self, request, response):
        self._step(self._app, request, response)

class EncryptPassword(Step):

    def __call__(self, app, request, response):
        import md5
        password = request.params['password']
        if password:
            request.params['encryptedPassword'] = md5.md5(password).hexdigest()
