#Copyright ReportLab Europe Ltd. 2000-2004
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/ers/test_dbutils.py
#$$

__version__=''' $Id'''
__doc__="""Test suite for database utilities.

To be completed - how can I connect? - uses reportlab.connect at present"""

import os, sys, time
from pprint import pprint as pp
import reportlab
from reportlab.test import unittest
from reportlab.test.utils import makeSuiteForClasses
from rlextra.ers.dbutils import Table, IntColumn, TextColumn, \
     DateTimeColumn, FloatColumn, ForeignKey, Index


import versionCheck

import mx.DateTime

class TestPerson(Table):
    "Here's one we will use for testing"
    _tableType = 'InnoDb'
    _fieldDefs = [
        IntColumn('id',primaryKey=1,allowNull=0),
        TextColumn('firstName',allowNull=0,size=50),
        TextColumn('lastName',allowNull=0,size=50),
        DateTimeColumn('dateOfBirth',allowNull=1)
        ]

class TestInvoice(Table):
    "Test something numeric"
    _tableType = 'InnoDb'
    _fieldDefs = [
        TextColumn('invCode',primaryKey=1,size=10,allowNull=0),
        DateTimeColumn('invDate',allowNull=0),
        TextColumn('customer',allowNull=0,size=50),
        FloatColumn('total',allowNull=0, default=0.0)
        ]

class TestInvoiceLine(Table):
    "Test something numeric"
    _tableType = 'InnoDb'
    _fieldDefs = [
        TextColumn('invCode',primaryKey=1,size=10,allowNull=0),
        IntColumn('lineNo',primaryKey=1,allowNull=0),
        TextColumn('item',size=50,allowNull=0),
        FloatColumn('amount',allowNull=0, default=0.0)
        ]
    # not the best place to put this but never mind for now
    _foreignKeys = [ForeignKey('invCode', 'TestInvoice','invCode')]



class DataObjectTests(unittest.TestCase):
    "Tests on simple database objects"

    def testKeyFields(self):
        self.assertEquals(TestPerson.getPrimaryKeyFieldNames(),
                          ['id'])
        fred = TestPerson(id=7,firstName='fred', lastName='bloggs')
        self.assertEquals(fred.getPrimaryKeyFieldNames(),['id'])
        self.assertEquals(fred.getPrimaryKeyValues(), [7])

        where, params = fred.getPrimaryKeyWhereClause()
        self.assertEquals(where, 'id = ?')

        f = ForeignKey('invCode', 'TestInvoice','invCode')
        self.assertEquals(f.getCreateClause(),
                          "FOREIGN KEY (`invCode`) REFERENCES `TestInvoice`(`invCode`)"
                          )

        f = ForeignKey(('invCode','lineNo'), 'TestInvoice',('invCode','lineNo'))
        self.assertEquals(f.getCreateClause(),
                          "FOREIGN KEY (`invCode`, `lineNo`) REFERENCES `TestInvoice`(`invCode`, `lineNo`)"
                          )

        
    def testIntColumn(self):        
        col = IntColumn('id')
        self.assertEqual(col.getCreateClause(),'`id` INTEGER NULL')
        col = IntColumn('prime', default=7, allowNull=0, primaryKey=1)
        self.assertEqual(col.getCreateClause(),'`prime` INTEGER NOT NULL DEFAULT 7')


        col = IntColumn('id', allowNull=0)
        self.assertEqual(col.assign('7'),7)
        self.assertEqual(col.assign(7.0),7)
        self.assertRaises(ValueError, lambda col: col.assign(None), col)

        col = IntColumn('id', allowNull=1)
        self.assertEqual(col.assign(None),None)

    def testTextColumn(self):
        col = TextColumn('name',size=50)
        
        self.assertEqual(col.getCreateClause(),'`name` VARCHAR(50) NULL')
        self.assertEqual(col.assign(None),None)
        #test string truncation
        self.assertEqual(col.assign('five-'*11),'five-'*10)

        #nulls
        col = TextColumn('name',size=50, allowNull=0)
        self.assertRaises(ValueError, lambda col: col.assign(None), col)
        
        #check single quote gets escaped        
        col = TextColumn('irishman',size=50,default="O'Grady")
        self.assertEqual(col.getCreateClause(),"`irishman` VARCHAR(50) NULL DEFAULT 'O''Grady'")

    def testDateTimeColumn(self):
        import mx.DateTime
        col = DateTimeColumn('dateOfBirth')
        self.assertEquals(col.getCreateClause(),'`dateOfBirth` DATETIME NULL')
        
        col = DateTimeColumn('dateOfBirth',default=mx.DateTime.DateTime(2003,11,17),allowNull=0)
        self.assertEquals(col.getCreateClause(),
                     "`dateOfBirth` DATETIME NOT NULL DEFAULT '2003-11-17 00:00:00.00'")

        col = DateTimeColumn('dateOfBirth',allowNull=0)
        self.assertRaises(ValueError, lambda col: col.assign(None), col)

        col = DateTimeColumn('dateOfBirth')
        self.assertEquals(col.assign(None), None)

        d = mx.DateTime.DateTime(2003,11,17)
        self.assertEquals(col.assign('2003-11-17'), d)
        #assert col.getCreateClause() == 'dateOfBirth DATETIME NULL'

    def testIndex(self):
        ind = Index('spam')
        self.assertEquals(ind.getCreateClause(), "INDEX `ind_spam` (`spam`)")

        ind = Index(['spam','eggs'])
        self.assertEquals(ind.getCreateClause(), "INDEX `ind_spam_eggs` (`spam`, `eggs`)")

        ind = Index(['spam','eggs'], 'myIndex')
        self.assertEquals(ind.getCreateClause(), "INDEX `myIndex` (`spam`, `eggs`)")

    def testTableInstance(self):
        fred = TestPerson()

        self.assertEquals(fred.getTableName(), 'TestPerson')
        self.assertEquals(fred.getPrimaryKeyFieldNames(), ['id'])
        

        #fields we expect to are defined and null
        self.assertEquals(fred.id, None)
        #self.assertEquals(fred.dateOfBirth, None)
        self.assertEquals(fred.firstName, None)
        self.assertEquals(fred.lastName, None)

        #fields we don't expect are not defined        
        self.assertRaises(AttributeError, lambda x: x.wrongField, fred)

        #we can set stuff and it will be checked
        fred.id = 32
        self.assertEquals(fred.id, 32)
        
        fred.id = 33.14
        self.assertEquals(fred.id, 33)

        fred.id = '34'
        self.assertEquals(fred.id, 34)

        fred.firstName = 'john'
        self.assertEquals(fred.firstName, 'john')
        
        fred.firstName = 90
        self.assertEquals(fred.firstName, '90')

        fred.firstName = 'longwinded' * 10
        self.assertEquals(fred.firstName, 'longwinded' * 5)

        self.assertRaises(ValueError, lambda bloke: setattr(bloke, 'firstName', None), fred)

        fred.dateOfBirth = '1966-03-21'
        d = mx.DateTime.DateTime(1966,3,21)
        self.assertEquals(fred.dateOfBirth, d)
        
        
    def testDatabaseSingleKey(self):
        
        conn = reportlab.connect('rptlab2_sandbox')

        TestPerson.createTable(conn, dropIfExists=1)
        

        #store new ones without an ID
        id = TestPerson(conn, firstName='Andy',lastName='Robinson', dateOfBirth='1966-03-21').store()
        TestPerson(conn, firstName='Alisa',lastName='Robinson', dateOfBirth='1965-09-12').store()
        TestPerson(conn, firstName='Tim',lastName='Robinson', dateOfBirth='1997-02-21').store()
        TestPerson(conn, firstName='Harry',lastName='Robinson', dateOfBirth='1999-06-25').store()

        #fetch one
        andy = TestPerson.fetchOne(conn, id)
        self.assertEquals(andy.firstName, 'Andy')
        self.assertEquals(andy.existsInDatabase(), 1)

        fred = TestPerson(conn)
        self.assertEquals(fred.existsInDatabase(), 0)
        
            
        #fetch bunch by keyword
        family = TestPerson.fetchMany(conn)
        self.assertEquals([p.firstName for p in family], ['Andy', 'Alisa','Tim','Harry'])

        #modify one
        andy.firstName = 'Ermintrude'
        id = andy.store()

        andy2 = TestPerson.fetchOne(conn, id)
        self.assertEquals(andy2.firstName, 'Ermintrude')

        #zap it
        andy2.remove()
        self.assertEquals(len(TestPerson.fetchMany(conn)), 3)

        cur = conn.cursor()
        cur.execute('DROP TABLE TestPerson')
        cur.close()
        
    def testDatabaseMultiKey(self):
        "Joint primary key and foreign key tests"
        conn = reportlab.connect('rptlab2_sandbox')

        TestInvoice.createTable(conn, dropIfExists=1)
        TestInvoiceLine.createTable(conn, dropIfExists=1)

        #store new ones without an ID
        tstamp = mx.DateTime.now()
        invId = TestInvoice(conn, invCode='XXX1234', invDate=tstamp, customer='goldlynch', total=1500).store()
        self.assertEquals(invId, 'XXX1234')

        invLine = TestInvoiceLine(conn,
                                 invCode=invId,
                                 lineNo=1,
                                 item='lifestyle consulting',
                                 amount=25000.0)
        lineId = invLine.store()
        #test update
        invLine.item = 'brain surgery'
        lineId2 = invLine.store()
        self.assertEquals(lineId, lineId2)  #same object


        invLine2 = TestInvoiceLine(conn,
                                 invCode='ABC1234',   #not there in foreign key, should raise error
                                 lineNo=1,
                                 item='no parent invoice',
                                 amount=25000.0)
        #not sure what error to trap
        try:
            invLine2.store()
            raise Exception("Should not have been able to store record due to foreign key constraints")
        except:
            pass
            
        #self.assertRaises(Exception, invLine2.store())
                          



        cur = conn.cursor()
        cur.execute('DROP TABLE TestInvoice')
        cur.execute('DROP TABLE TestInvoiceLine')
        cur.close()
        
        
    def testBatchInsert(self):
        
        conn = reportlab.connect('rptlab2_sandbox')
        TestInvoice.createTable(conn, dropIfExists=1)

        fieldNames = ['invCode','invDate', 'customer','total']
        data = [('ABC0001','2003-09-17','HP','1000.00'),
                ('ABC0002','2003-09-18','IBM','2000.00'),
                ('ABC0003','2003-09-19','ADOBE','3000.00'),
                ('ABC0004','2003-09-20','SYBASE','4000.00'),
                ('ABC0005','2003-09-21','ORACLE','5000.00')
                ]

        TestInvoice.insertRows(conn, fieldNames, data)

        objects = TestInvoice.fetchMany(conn)
        self.assertEquals(len(objects), len(data))

                                   
        cur = conn.cursor()
        cur.execute('DROP TABLE TestInvoice')
        cur.close()
        
        
def makeSuite():
    return makeSuiteForClasses(DataObjectTests)


#noruntests
if __name__ == "__main__":
    unittest.TextTestRunner().run(makeSuite())
