
__version__=''' $Id'''
__doc__="""Test ErsWebApps over HTTP."""


import unittest

class ErsTests(unittest.TestCase):

    url = None  # This must be set to URL to test
    document = 'NOTSET' # This must be set to test non-webapp ERS apps

    class Response:

        def __init__(self):
            content = None
            info = None

    def runRequest(self, params = {}, urlJoin = None):
        url = self.url
        if urlJoin:
            import urlparse 
            url = urlparse.urljoin(url, urlJoin)
        response = self.Response()
         # Use urllib2.urlopen (not urllib) because it raises exceptions on HTTP errors
        import urllib2, urllib
        if params:
            query = urllib.urlencode(params)
            f = urllib2.urlopen(url, query)
        else:
            f = urllib2.urlopen(url)
        response.content = f.read()
        response.info = f.info()
        f.close()
        reportedContentLength = response.info.getheader('Content-Length', None)
        if reportedContentLength is not None:
            reportedContentLength = int(reportedContentLength)
            receivedContentLength = len(response.content)
            self.failUnless(
                reportedContentLength == receivedContentLength,
                'Content length incorrect (reported %d, received %d)' % (reportedContentLength, receivedContentLength)
            )
        return response

    def runErsRequest(self, params):
        reqParams = {
            'document': self.document,
            'ers_action': '',
            'validating': 'yes',
            'outfile': '',
            'datafilters': '',
        }
        reqParams.update(params)
        return self.runRequest(reqParams)

    def getLinkHRefs(self, html):
        "Return list of href values extracted from any <a> tags in html"
        import re
        return re.findall(r"""<a\s+href\s*=\s*["']?([^ >'"]+)""", html)


class HttpTests(ErsTests):

    # This must be set to filename of XML sample file
    sampleFilename = None

    def setUp(self):
        assert self.sampleFilename != None, 'No sample XML filename available'
        self.sampleData = open(self.sampleFilename, 'rb').read() 

    def tearDown(self):
        self.sampleData = None

    def testRml(self):

        # Get RML for sample
        sampleData = self.sampleData
        params = {
            'action': 'Rml',
            'data': sampleData,
        }
        response = self.runErsRequest(params)

        # Check response okay 
        contentLength = len(response.content)
        contentType = response.info.gettype() 
        self.failUnless(contentLength > 20) # Check there's something in the content!
        # RML should be escaped in HTML, check content type
        self.failUnless(contentType == 'text/html')

    def testPlainRml(self):

        sampleData = self.sampleData
        params = {
            'action': 'PlainRml',
            'data': sampleData,
        }
        response = self.runErsRequest(params)

        # Check response okay 
        contentLength = len(response.content)
        contentType = response.info.gettype() 
        self.failUnless(contentLength > 20) # Check there's something in the content!
        self.failUnless(contentType == 'data/x-rml', 'Incorrect content-type: %s' % contentType)

    def testPdf(self):

        maxSizeDifference = 100
        minPdfSize = 100000

        # Get PdfText for sample
        sampleData = self.sampleData
        params = {
            'action': 'PdfText',
            'data': sampleData,
        }
        response = self.runErsRequest(params)

        # Check response okay 
        contentLength = len(response.content)
        contentType = response.info.gettype() 
        self.failUnless(contentLength > minPdfSize)
        self.failUnless(contentType == 'application/pdf')
        
        # Get Gooped version of sample
        params = {
            'action': 'GoopHTML',
            'data': sampleData,
        }
        goop = self.runErsRequest(params).content

        # Get PdfText for gooped sample
        params = {
            'action': 'PdfText',
            'data': goop,
            'datafilters': 'unbase64,flate',
        }
        goopResponse = self.runErsRequest(params)

        # Check response valid and similar to response from non-gooped version
        # Just do crude check to see if both PDFs are similar size here,
        # really want to be able to compare PDFS byte-for-byte but need the PDF invariant
        # stuff to work correctly for this.
        sizeDiff = abs(len(goopResponse.content) - contentLength)
        self.failIf(sizeDiff > maxSizeDifference , "Suspicious: Gooped and non-gooped PDFs differ in size by %d" % sizeDiff)
        self.failUnless(goopResponse.info.gettype() == contentType)

        # Get PdfLink for sample
        params = {
            'action': 'PdfLink',
            'data': sampleData,
        }
        response = self.runErsRequest(params)
        self.failUnless(response.info.gettype() == 'text/html')

        # extract PDF link from resulting HTML
        pdfHRefs = [href for href in self.getLinkHRefs(response.content) if href.lower().endswith('.pdf')]
        self.failUnless(len(pdfHRefs) > 0)

        # Download linked PDF
        response = self.runRequest(urlJoin = pdfHRefs[0])

        # Check response valid and PDF similar to response from PDFText
        sizeDiff = abs(len(response.content) - contentLength)
        self.failIf(sizeDiff > maxSizeDifference , "Suspicious: PDF from PDFLink and PDF from PDFText differ in size by %d" % sizeDiff)
        self.failUnless(response.info.gettype() == contentType)

    
def runTests(testCases, url):
    suite = unittest.TestSuite()
    for test in testCases:
        test.url = url
        suite.addTest(unittest.makeSuite(test))
    unittest.TextTestRunner(verbosity=2).run(suite)


import getopt, sys, os
def handleCommandLine(testCases):

    """ Call this from your test script passing in a list of testCase classes
        to handle running your tests from the command-line.
    """

    def usage(msg = None):
        progName = os.path.basename(sys.argv[0])
        print
        if msg:
            print 'ERROR: %s\n' % msg
        print 'Usage:'
        print '%s --url=<URL to test>' % progName
        
    try:
        opts, args = getopt.getopt(
            sys.argv[1:], 'hu:', ['help', 'url=']
        )
    except getopt.GetoptError:
        # print help information and exit:
        usage('Invalid syntax')
        sys.exit(2)

    url = None
    for o, a in opts:
        if o in ('-h', '--help'):
            usage()
            sys.exit()
        elif o in ('-u', '--url'):
            url = a
    if not url:
        usage('Must specify URL to test (e.g. http://server.reportlab.com/cgi-bin/webeforms.cgi)') 
        sys.exit(2)

    runTests(testCases, url)


if __name__ == '__main__':
    raise RuntimeError, 'No runnable tests here, run tests in specific ERS apps instead'
