#Copyright ReportLab Europe Ltd. 2000-2004
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/ers/testwebapp.py

__version__=''' $Id'''
__doc__="""Beginning of a test suite for web apps.

No real test strategy yet..."""

import os, sys, time
from pprint import pprint as pp
from reportlab.test import unittest
from reportlab.test.utils import makeSuiteForClasses
from rlextra.ers.webapp import WebApp, WebRequest, WebResponse
from rlextra.utils.checkhtml import PageChecker




class WebAppTestCase(unittest.TestCase):
    "Tests on simple web app components"

    def testResponse(self):
        resp = WebResponse()
        resp.setContentType('text/plain')
        resp.write('This is a rather simple example:\n')
        resp.write('Just two lines of text\n')
        resp.getOutput()
        

    def testDefaultAction(self):
        req = WebRequest()
        resp = WebApp().handleRequest(req)
        assert resp.containsText('Server alive')
        assert resp.containsText('COMPLETED:SUCCESSFULLY')
        
    def testPing(self):
        req = WebRequest(action='ping')
        resp = WebApp().handleRequest(req)
        assert resp.containsText('pong')


    def testAuthenticationModes(self):
        "Thin tests of the authentication state machine"
        app = WebApp(authenticating=1,
                     authDummyUsers=1,
                     authTimeOut=0.1
                     )


        # no user, no token, no lsp
        req = WebRequest()
        app.authenticate(req)
        assert req.user == None
        assert req.params['action'] == 'showLogin'
        assert req.params['ERROR_MESSAGE'] == "Login Required"

        #username/password OK
        req = WebRequest(USER='alice',PASSWORD='alice', action='ping')
        app.authenticate(req)
        assert req.user == 'alice'
        assert req.params['action'] == 'ping'
        assert req.params.has_key('token')
        goodToken = req.params['token']  #use this lower down
        
        #username/password bad
        req = WebRequest(USER='alice',PASSWORD='alicia', action='ping')
        app.authenticate(req)
        assert req.user == 'alice'
        assert req.params['action'] == 'showLogin'
        assert req.params.has_key('login_saved_params')
        assert req.params['ERROR_MESSAGE'] == 'Bad username and/or password'

        #good fresh token
        req = WebRequest(token=goodToken, action='ping')
        app.authenticate(req)
        assert req.user == 'alice'
        assert req.params['action'] == 'ping'
        assert req.params.has_key('token')

                
        #stale token
        time.sleep(0.15)
        req = WebRequest(token=goodToken, action='ping',extra='spurious')
        app.authenticate(req)
        assert req.user == 'alice'
        assert req.params['action'] == 'showLogin'
        assert req.params.has_key('login_saved_params')
        assert req.params['ERROR_MESSAGE'] == 'Timed out, please log in again'
        goodLsp = req.params['login_saved_params']
        
        #get a good token again
        req = WebRequest(USER='alice',PASSWORD='alice', action='ping')
        app.authenticate(req)
        goodToken = req.params['token']  #use this lower down

        #good token, user and password - user and pass should be chucked 
        req = WebRequest(USER='alice',PASSWORD='alice', action='ping', token=goodToken)
        app.authenticate(req)
        assert not req.params.has_key('USER')
        assert not req.params.has_key('PASSWORD')
        assert req.user == 'alice'
        assert req.params['action'] == 'ping'
        assert req.params.has_key('token')

        #4. lsp only - should not happen        
        req = WebRequest(action='ping', login_saved_params=goodLsp)
        app.authenticate(req)
        assert req.user == None
        assert req.params['action'] == 'showLogin'
        assert req.params['ERROR_MESSAGE'] == "Login Required"
        
        #5. genuine login or relogin being submitted
        req = WebRequest(USER='alice',PASSWORD='alice', login_saved_params=goodLsp)
        app.authenticate(req)
        assert req.user == 'alice'
        assert req.params['action'] == 'ping'
        assert req.params['extra'] == 'spurious'  #this was in the lsp
        assert req.params.has_key('token')
        

        #6 spoof detection - time out Alice, try to get in as Bob
        req = WebRequest(USER='bob',PASSWORD='bob', login_saved_params=goodLsp)
        app.authenticate(req)
        assert req.user == 'alice'
        assert req.params['action'] == 'showLogin'
        assert req.params['ERROR_MESSAGE'][0:11] == "You are not"
        

    def testAuthenticationKicksIn(self):
        # the dummy gatekeeper doe snot use a database
        # and support predefined user/password pairs
        # 'alice/alice','bob/bob' etc.
        app = WebApp(authenticating=1,
                     authDummyUsers=1
                     )
        
        req = WebRequest()
        req.params = {'action':'ping'}
        resp = app.handleRequest(req)
        #shows we got bounced to the login page
        assert resp.containsText('ReportLab Login')

        #now find a field or two I expect to be there
        checked = PageChecker(resp.getOutput())
        assert checked.hasFormField('login_saved_params')

        #decode saved stuff form the original request
        lsp = checked.getFormFieldValue('login_saved_params')
        import cPickle, base64
        lspDict = cPickle.loads(base64.decodestring(lsp))
        assert lspDict['action'] == 'ping'
        
        #now try a login with a password
        req = WebRequest()
        req.params = {'USER':'alice',
                      'PASSWORD':'alice',
                      'login_saved_params':lsp}
        resp = app.handleRequest(req)
        #expect to get past the login page to where we wanted to be
        assert resp.containsText('pong')

    def testTimeOut(self):
        "Log in, wait"
        app = WebApp(authenticating=1,
                     authDummyUsers=1,
                     authTimeOut=0.1
                     )
        
        
        req = WebRequest()
        req.params = {'action':'testForm'}
        resp = app.handleRequest(req)
        assert req.user is None
        #shows we got bounced to the login page
        assert resp.containsText('ReportLab Login')

        #now find a field or two I expect to be there
        checked = PageChecker(resp.getOutput())
        assert checked.hasFormField('login_saved_params')

        #decode saved stuff form the original request
        lsp = checked.getFormFieldValue('login_saved_params')
        import cPickle, base64
        lspDict = cPickle.loads(base64.decodestring(lsp))
        assert lspDict['action'] == 'testForm'
        
        #now try a login with a password
        req = WebRequest()
        req.params = {'USER':'alice',
                      'PASSWORD':'alice',
                      'login_saved_params':lsp}
        resp = app.handleRequest(req)
        assert req.user == 'alice'
        #expect to get past the login page to where we wanted to be:
        #the web form
        assert resp.containsText('testForm')

        #extract the token
        checked = PageChecker(resp.getOutput())
        token = checked.getFormFieldValue('token')

        #verify a repeat login with the token works
        req = WebRequest()
        req.params = {'token':token,
                      'action':'testForm',
                      'textfield':'aaa'
                      }
        resp = app.handleRequest(req)
        assert resp.containsText("Old value of textfield was 'aaa'")
        assert req.user == 'alice'

        #verify a short wait times you out
        import time
        time.sleep(0.15)
        req = WebRequest()
        req.params = {'token':token,
                      'action':'testForm',
                      'textfield':'aaa'
                      }
        resp = app.handleRequest(req)
        assert resp.containsText("ReportLab Login")


    def testSnip(self):
        from rlextra.ers.webapp import _substituteSnip
        text1 = "Before<!--middlesnip-->BLAHBLAH<!--middlesnip-->After"
        out = _substituteSnip(text1, middle='During')
        #print '\noutput=',out
        self.assertEquals(out, "BeforeDuringAfter")

        text1 = "Before<!--middlesnip-->BLAHBLAH<!--middlesnip-->Aft<!--Asnip-->BLAH<!--Asnip-->er"
        out = _substituteSnip(text1, middle='During', A='')
        self.assertEquals(out, "BeforeDuringAfter")

        text1 = "Before<!--middlesnip-->BLAHBLAH<!--middlesnip-->Aft<!--Asnip-->BLAH<!--Asnip-->er<!--Bsnip-->Alone"
        out = _substituteSnip(text1, middle='During', A='', B='ZZZZZZZ')
        self.assertEquals(out, "BeforeDuringAfter<!--Bsnip-->Alone")

        #do multiple replaces work?
        input = "Before<!--middlesnip-->BLAHBLAH<!--middlesnip-->After" * 3
        out = _substituteSnip(input, middle='During')
        self.assertEquals(out, "BeforeDuringAfter" * 3)

        
def makeSuite():
    return makeSuiteForClasses(WebAppTestCase)


#noruntests
if __name__ == "__main__":
    unittest.TextTestRunner().run(makeSuite())