#webapp.py - not done yet but the docstring is kind of
#a spec - 7/3/2002

"""This provides a base class for web applications.

After some study of JSP, ASP, Twisted, Webware etc
I realised our standard controllers are oversimplifying
things.  There is a 'standard model' used by all these
established frameworks which is just complex enough to
capture what goes on.

WebApp holds any 'state' for the app and handles requests.
It should be written so it will work OK in a long running
framework and be able to handle multiple concurrent requests.
It will provide a number of methods to be overridden which
allow pre- or post-processing of requests.

WebRequest holds any info passed into the request.
For now it has an attribute 'params' holding any parameters
in a dictionary, but it could also hold a lot more stuff
abstracted from the CGI or server environment e.g. calling
IP address, browser details.

WebResponse handles the output. It's just a buffer you
write HTML or other data into, but you can change its
content-type from HTML to GIF or whatever, right up to
the end.  It is very important to write to a response
and not just use 'print', as (a) this lets us write
test cases which work without going through the web and
(b) if ever we switch an app to a long running server
then print statements will be a risk and require careful
synchronization.

RlWebApp is a subclass of WebApp which has some common
knowledge about things ReportLab wants to standardise in
its web applications: logging, completion marks, user
management.  It will have knowledge of preppy views, and
of QDF files as they mature.
1. we try for a method matching the action parameter
2. we try for a prep file in the views subdirectory

A major goal is that web apps should be as fully testable
as possible from scripts without needing a browser or web
server; one should be able to prepare request/response
assertions and test them in isolation.

DEBUGGING CGI support
    1) add SAVE_LASTEST_PARAMS='pathparams.txt'
    to the keyword arguments.
    2) set PSEUDO_CGI=yes in command line env
    3) pythonwin xxx.cgi

TODO
- factor out user validation
- argue about font paths
- check you can't re-login as wrong user
- conventions/conveniences for prep file authors
  to always get the token included
- error handling conventions
- LIVE mode

"""

import sys, os, re
import base64, cPickle, md5, urllib
from types import ListType, TupleType, StringType
from rlextra.preppy import preppy
from rlextra.radxml.xmlutils import translateStandardEntities, quoteValue

COMPLETE_UNSET = "***NO COMPLETION CODE SET***"
COMPLETION_START = "<!-- COMPLETION CODE: "
COMPLETION_END = " END OF COMPLETION CODE -->"
COMPLETE_SUCCESS = "Completed Normally"
COMPLETE_USER_ERROR = "User input parameters problem"
COMPLETE_LOGIC_ERROR = "System Logic Problem (unexpected error)"
COMPLETE_SECURITY_ERROR = "Attempted Security Violation (this has been logged)"

try:
    import versionCheck
except ImportError:
    pass

def getCgiParams(keep_blank_values=0, quoteFunc=None):
    "returns dictionary of parameters found in the cgi script"
    if quoteFunc is None:
        quoteFunc = quoteValue
    dictionary = {}
    import cgi
    form = cgi.FieldStorage(keep_blank_values=keep_blank_values)
    #dbg = open('/tmp/dbg.txt','w')
    for name in form.keys():
        if name.endswith(':file'):
            value = form[name]
            name = name[:-5]
        else:
            value = form.getvalue(name)
            if name.endswith(':list'):
                name = name[:-5]
                if isinstance(value, ListType):
                    value = map(quoteFunc, value) 
                else:
                    value = [quoteFunc(value)]
            else:
                assert not isinstance(value, ListType), 'Received multiple values for CGI param "%s"' % name
                value = quoteFunc(value)
        dictionary[name] = value
    return dictionary

def getCommandLineParams():
    "return any key=value params in command line as a dict"
    params = {}
    for arg in sys.argv:
        chunks = arg.split('=')
        if len(chunks) == 2:
            params[chunks[0]]=chunks[1]
    return params


def _substituteSnip(html, **kw):
    """A light substitution system used in doHeader/doFooter.
    Works on pure HTML files.
    If you supply a key "title='My Page'", it will look for
    2 successive HTML comments
        <!--titlesnip-->Blah Blah<!--titlesnip-->
    and replace them with your content.

    """
    for (key, value) in kw.items():
        snip = '<!--%ssnip-->' % key
        found1 = html.find(snip)
        while found1 > -1:
            found2 = html.find(snip, found1 + len(snip))
            if found2 == -1:
                break
            endPos = found2 + len(snip)
            firstBit = html[0:found1]
            lastBit = html[endPos:]
            html = firstBit + value + lastBit
            found1 = html.find(snip, len(firstBit) + len(value))  #there might be more...
    return html

def _getRotor():
    "Basic, standard scrambler for obfuscating parameters"
    import rotor
    return rotor.newrotor('bingo-bongo')

def _get_remote_addr():
    from os import environ
    return environ.get('REMOTE_ADDR','[no-remote-addr]').replace('-','_')

class WebRequest:
    def __init__(self, **kw):
        #expand this to put in all the things potentially retrievable
        #from the request - not just the parameters dictionary
        self.url = None
        self.user = None
        self.params = {}

        #allows keywords passing in for convenience in writing test cases
        if kw.has_key('url'):
            self.url = kw['url']
            del kw['url']

        self.params.update(kw)

    def isSSL(self):
        """ Return true if we are on a SSL (https) connection. """
        return os.environ.get('SSL_PROTOCOL', '') != ''

    # modified from recipe in Python Cookbook...
    def getScriptURL(self):
        """ Return a full URL starting with schema, servername and port.
            *uri* -- append this server-rooted uri (must start with a slash)
        """
        schema, stdport = (('http', '80'), ('https', '443'))[self.isSSL()]
        host = os.environ.get('HTTP_HOST', None)
        if not host:
            host = os.environ.get('SERVER_NAME')
            port = os.environ.get('SERVER_PORT', '80')
            if port != stdport: host = host + ":" + port
        path = self.url
        if path is None:
            path = os.environ.get('SCRIPT_NAME')
        return "%s://%s%s" % (schema, host, path)

    def makeGetUrl(self, **params):
        """Easy way to make a get request back to same URL - use this
        in prep files"""
        if params:
            return self.url + '?' + urllib.urlencode(params)
        else:
            return self.url


class WebResponse:
    def __init__(self):
        self._headers = {}
        self._outputList = []
        self.trailers = []
        self.setContentType('text/html') # unless overridden

    def setHeader(self, key, value):
        self._headers[key] = value

    def setContentType(self, typeCode):
        self._headers['Content-type'] = typeCode

    def getContentType(self):
        return self._headers['Content-type'].strip()

    def write(self, stuff):
        "Make it look like a file object"
        self._outputList.append(stuff)

    def getContent(self):
        T = self._outputList
        n = len(T)
        a = T.append
        for t in self.trailers:
            if callable(t): t = t()
            a(t)
        s = ''.join(T)
        del self._outputList[n:]
        return s

    def getOutput(self):
        "Return the full HTTP response as a string"
        buf = []
        content = self.getContent()
        length = len(content)

        # conventionally content-length comes after
        # content-type; for others we do not care
        # about the order

        ctype = self._headers['Content-type']
        try:
            del self._headers['Content-type']
            buf.append('Content-type: %s\r\nContent-length: %d\r\n' % (ctype, length))

            for (key, value) in self._headers.items():
                buf.append('%s: %s\r\n' % (key, value))
        finally:
            self.setContentType(ctype)
        buf.append('\r\n')
        buf.append(content)
        return ''.join(buf)

    def respond(self, device=sys.stdout):
        """What it does at the end.

        You can hook into this more easily by testing the value
        of getOutput"""
        device.write(self.getOutput())

    def containsText(self, text, justContent=None):
        "Use for writing test cases"
        found = (justContent and self.getOutput or self.getContent)().find(text)
        return found > -1

    def containsPattern(self, pattern):
        "Support for test cases.  Make a regex from pattern"
        pass

    def getHeader(self, headerType):
        "Support for test cases.  Get value of header, or raise key error"
        pass

def include(viewName, dictionary={}, **kwdict):
    "Simple include utility for preppy."
    remove = []
    old = {}
    for k, v in kwdict.items():
        if dictionary.has_key(k):
            old[k] = dictionary[k]
        else:
            remove.append(k)
        dictionary[k] = v
    dir, filename = os.path.split(viewName)
    if dictionary.get('_PROJECT_RELATIVE_INCLUDE',1) and  not os.path.isabs(dir): dir = os.path.join(dictionary.get('APPDIR','.'),dir)
    root, ext = os.path.splitext(filename)
    m = preppy.getModule(root, directory=dir, source_extension=ext)
    outputList = []
    m.run(dictionary, __write__= outputList.append)
    text = ''.join(outputList)
    # now restore dictionary
    for k in remove:
        del dictionary[k]
    for k,v in old.items():
        dictionary[k] = old[k]
    return text

class WebApp:
    """Basic dispatch mechanism for web applications.
    You can either set attributes or subclass this to get different
    behaviour.
    """
    _uid = ''
    _needCompletionMark_re = re.compile('^.*/(?:html|xml)(?:;.*|)$',re.I)   #match content types that need completion mark
    def __init__(self,**kw):
        import errhandlers
        import apploggers
        self._setKeywords(**kw)         # set things we are given
        self._setKeywords(              # set our defaults
                OUTDIR = None,   #your CGI should set this if it makes files
                HTTP_OUTDIR = None, #..and this too.
                fontPath = ['../fonts','./fonts','..','.'], #extra directories to search for T1 fonts
                                #which we need to make bitmap charts.  DISCUSS (A)

                actionSynonyms = ['action','ERS_ACTION'],
                #actionMethodPrefix = 'action_',  # leave configurable or delete?
                viewDir = 'views',
                defaultAction = 'alive',
                dbName = None,
                #preppyFriendly = 1,     #remove?
                #cgiName = None,  #per request? duplicate?

                #html helpers
                pageTemplate=None,


                # these are for optional database support
                _conn = None,
                _cur = None,            #optional db support

                # these are used if in authenticating mode.
                authenticating = 0,  #turn on or off.
                authTimeOut = 600,
                authDummyUsers = 0,  #substitutes 4 dummy users
                # so we can try authenticating apps without
                # a database
                authLoginTable = 'WebPasswords',  #where to look
                authUserCol = 'PersonId',
                authPasswordCol = 'EncryptedPassword',
                # Callers can override this to install
                # a different error handler
                errorHandler = errhandlers.ShowTracebackHandler(),
                logger = apploggers.AppLogger,
                LIVE=0,
                _completion_mark = None,    #dangling health indicator
                )
        #(A) - Robin says have this always on
        if self.fontPath:
            from reportlab.rl_config import T1SearchPath
            for d in self.fontPath:
                d = os.path.abspath(d)
                if d not in T1SearchPath and os.path.isdir(d):
                    T1SearchPath.insert(0,d)


        # some internals that should be initialized
        self._pageHeader = None
        self._pageFooter = None

        #start up our logger; this will provide our ID and start time
        if self.logger and callable(self.logger): self.logger = self.logger()

    def _setKeywords(self,**kw):
        for k,v in kw.items():
            if not self.__dict__.has_key(k):
                setattr(self,k,v)

    def handleCgi(self):
        "Easy entry point for CGI scripts, does everything"
        write = sys.stdout.write
        logger = self.logger
        try:
            req = WebRequest()
            req.params = params = {}
            save_lastest_params= getattr(self,'SAVE_LASTEST_PARAMS',None)
            if getattr(self,'PSEUDO_CGI',os.environ.get('PSEUDO_CGI',None)) not in (1,'1','yes','y'):
                req.url = params['CGI_SCRIPT'] = params['SCRIPT_NAME'] = os.environ['SCRIPT_NAME']
                params.update(getCgiParams(
                    keep_blank_values=int(getattr(self,'keep_blank_values','0')),
                    quoteFunc=getattr(self,'paramQuoteFunction', None),
                ))
                try:
                    from rlextra.utils.buildutils import safe_makedirs
                    safe_makedirs(self.OUTDIR,0777)
                    if save_lastest_params:
                        f = open(save_lastest_params,'w')
                        f.write(repr(params))
                        f.close()
                        del f
                except:
                    pass
            else:
                params.update(eval(open(save_lastest_params,'r').read()))
                req.url = params['CGI_SCRIPT']

            logger.record('startcgi',params=params)
            resp = self.handleRequest(req)
            write(resp.getOutput())
            logger.record('finish',headers=resp._headers)
        except:
            if not self.errorHandler: raise
            et, ev = map(str,sys.exc_info()[:2])
            ftb = self.errorHandler.handleException()
            logger.record('error',ftb=ftb,error_type=et,error_value=ev)
            resp = WebResponse()
            params['FORMATTEDTRACEBACK'] = ftb
            params['live'] = live = str(int(self.LIVE or params.get('live','0')=='1'))
            if not params.has_key('ERROR_MESSAGE'):
                    params['ERROR_MESSAGE'] = 'An unexpected error occurred'
            from xml.sax.saxutils import escape
            params['ERROR_VALUE'] = ev = escape(ev)
            params['ERROR_TYPE'] = et = escape(et)
            params['doHeader'] = self.doHeader
            params['doFooter'] = self.doFooter
            try:
                self.action_showError(req,resp)
            except:
                if live=='1': ftb = '%s: %s' % (et,ev)
                a = resp.write
                a('<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">')
                a('<html><head><title>Error</title></head><body>')
                a('<h1>Error</h1>')
                a('<p>The server encountered an internal error or misconfiguration ')
                a('and was unable to complete your request.</p>')
                a(ftb)
                a('</body></html>')
            resp.setHeader('Content-type','text/html')
            resp.setHeader('Status','500 Internal Server Error')
            if not self._completion_mark: self.setLogicErrorCompletionMark()
            if self.needCompletionMark(req,resp): self.applyCompletionMark(req, resp)
            write(resp.getOutput())
        logger.store()

    def _handleDict(self, D, url='dict'):
        "handle a request constructed from dictionary D"
        req = WebRequest()
        req.url = url
        req.params = D
        return self.handleRequest(req)

    def handleDict(self, D, url='dict'):
        sys.stdout.write(self._handleDict(D,url=url).getOutput())

    def handleCommandLine(self):
        "Easy entry point for command line testing"
        self.handleDict(getCommandLineParams(),'commandline')

    def handleRequest(self, request):
        """Accept a WebRequest, return a WebResponse.
        Ensure every request goes through this, whether invoked
        from CGI or command line or anything else."""
        response = WebResponse()
        return self._handleRequest(request,response)

    def dispatch(self,request,response):
        '''get the action name and try to run it'''
        self._actionName = actionName = self.getActionName(request)
        # first try at dispatch:  look for a method
        methodName = 'action_' + actionName
        method =  getattr(self, methodName, None)
        if method is not None:
            method(request, response)
        else:
            self.runView(request, response, actionName)

    def _handleRequest(self, request, response):
        #0. before hook
        #1. any security checking / authentication and lookup
        # of user details (LoginManager?)

        self.authenticate(request)
        self.dispatch(request,response)
        if self.needCompletionMark(request,response): self.applyCompletionMark(request, response)
        return response

    def _getPreppyModOutput(self,view,viewDir,request,response,**kw):
            mod = preppy.getModule(view,viewDir,savePyc=0)
            D = {'controller': self, 'app': self, 'request':request, 'response':response}
            def myInclude(viewPath,dictionary=D,useCopy=None,**kw):
                if useCopy: dictionary=dictionary.copy()
                return include(viewPath,dictionary=dictionary,**kw)
            D['include'] = myInclude
            D.update(request.params)
            D.update(kw)
            return mod.getOutput(D)

    def runView(self, request, response, actionName, xtraDict = {}, raiseError = 1):
        """Run the named view, look for it in standard locations.
        Raise error if view not found and raiseError=1. Return true if successful
        """
        viewOk = self._runView(request, response, actionName, xtraDict = xtraDict) \
            or self._runView(request, response, actionName, defDirs=1, xtraDict = xtraDict)
        if (not viewOk) and raiseError:
            raise ValueError, 'Unknown view/action "%s"' % actionName
        return viewOk

    def _runView(self, request, response, actionName, defDirs=0, xtraDict={}):
        viewName = self.viewDir + os.sep + actionName
        if defDirs:
##            try:  AR this doesn't work, __file__ not defined
##                viewName = os.path.dirname(__file__) + os.sep + 'html' + os.sep + actionName
##            except:
            from rlextra.ers import webapp
            viewName = os.path.dirname(webapp.__file__) + os.sep + 'html' + os.sep + actionName
        from reportlab.lib.utils import rl_isfile
        OK = rl_isfile(viewName + '.pyc') or rl_isfile(viewName + '.prep')
        if OK:
             response.write(self._getPreppyModOutput(os.path.basename(viewName),os.path.dirname(viewName),request,response,**xtraDict))
        return OK

    def getUniqueFileName(self, prefix=None,suffix=None,ext='.pdf',hostname=None):
        # builds one from timestamp
        try:
            id = self.logger.id
            appName = self.logger.appName
        except:
            from rlextra.utils.cgisupport import TimeStamp
            id = TimeStamp().currentId
            appName = 'out-'
        if not prefix: prefix = appName 
        if prefix and prefix[-1]!='-': prefix += '-'
        if not hasattr(self,'_fnc'): self._fnc = 0
        else: self._fnc = self._fnc+1
        if hostname:
            if type(hostname) is not type(''):
                import socket
                hostname = socket.gethostname()
            prefix += hostname + '-'
        id += str(self._fnc)
        if suffix: id += ('-'+suffix)
        return prefix+id+ext

    def handleError(self, request, response, error):
        response.setContentType('text/html')
        response.write('<pre>An error occurred...')
        raise

    def getActionName(self, request):
        "Work out the name of the main action. Override if needed"
        for key in self.actionSynonyms:
            try:
                return request.params[key]
            except KeyError:
                pass
        # if no action given, default is expected
        return self.defaultAction

    def needCompletionMark(self,request,response):
        return self._needCompletionMark_re.match(response.getContentType()) is not None

    def applyCompletionMark(self, request, response):
        "Tiny page footer which is easy to parse"
        ctype = self._completion_mark or COMPLETE_UNSET
        info = (ctype,)+getattr(self,'_completion_args',())
        response.write("\n%s\n%s\n%s\n" % (COMPLETION_START, repr(info), COMPLETION_END))

    def setCompletionMark(self,tag,*args):
        if not args: args = (getattr(self,'_actionName','unknown_action'),)
        self._completion_mark = tag
        self._completion_args = args

    def setSuccessCompletionMark(self,*args):
        self.setCompletionMark(COMPLETE_SUCCESS,*args)

    def setUserErrorCompletionMark(self,*args):
        self.setCompletionMark(COMPLETE_USER_ERROR,*args)

    def setLogicErrorCompletionMark(self,*args):
        self.setCompletionMark(COMPLETE_LOGIC_ERROR,*args)

    def setSecurityErrorCompletionMark(self,*args):
        self.setCompletionMark("Attempted Security Violation (this has been logged)",*args)

    def error_if_live(self, response, contentType="text/html", request=None):
        if not (self.LIVE or getattr(request,'params',{}).get('live','0')=='1'):
            response.setContentType(contentType)
        else:
            class WebAppSecurityError(Exception):
                """Raise these for attempts to call development stuff in LIVE"""
                pass
            if not self._completion_mark: self.setSecurityErrorCompletionMark()
            raise WebAppSecurityError

    ###################################################################
    #
    #   readymade actions you can test
    #
    ###################################################################

    def action_alive(self, request, response):
        response.write("<p>Server alive, no action specified!<p>")

    def action_echo(self, request, response):
        response.write('<html><head><title>Parameter Dump</title></head><body>'
                       '<p>Parameters:</p><pre>')
        for param, value in request.params.items():
            response.write('%s: %r\n' % (param, value))
        response.write('</pre></body></html>')

    def action_ping(self, request, response):
        "Test function always there"
        response.setContentType('text/plain')
        response.write('pong\n')

    def action_tooManyLogins(self, request, response):
        response.write("<p>Too many login attempts.  Goodbye.<p>")

    def action_showLogin(self, request, response, xtraDict={}):
        "This one may be called by the authenticator"
        actionName = 'login'
        self.runView(request, response, actionName, xtraDict=xtraDict)

    def action_showError(self, request, response, xtraDict={}):
        "This one will be called when errors occur"
        actionName = 'error'
        if not self.runView(request, response, actionName, xtraDict=xtraDict, raiseError=0):
            raise ValueError('No view named %s found!' % actionName)

    def action_testForm(self, request, response):
        "This is a simple form that can be resubmitted to test tokens/authent"
        user = request.user
        timeout = self.authTimeOut
        token = request.params.get('token','')
        textfield = request.params.get('textfield','')
        response.write(self.doHeader(title="Authentication Test Form"))
        response.write("""
        <h2>Authentication test form</h2>
        <p>This form lets us test the basic mechanisms - can we
        submit back to the same CGI and carry the token across? - and
        thereby lets us write tests for authentication, timeout and so on.</p>
        <p>user = %(user)s, timeout = %(timeout)s</p>
        <hr>
        <p>Old value of textfield was '%(textfield)s'</p>
        <form name="form1" method="post">
        <input type='hidden' name='token' value='%(token)s'>
        <p><input type="text" name="textfield" value='%(textfield)s'></p>
        <p><input type="submit" name="action" value="testForm"></p>
        </form>
        <hr>

        """%locals())
        response.write(self.doFooter())
        return response


    ###################################################################
    #
    #   helpers for HTML page development
    #
    ###################################################################


    def _initPageTemplate(self):
        "Called on first request do doHeader/doFooter."
        if self.pageTemplate is None:
            #use the checked in one in rlextra/ers
            import rlextra.ers
            pageTemplate = os.path.dirname(rlextra.ers.__file__) + os.sep + 'html' + os.sep + 'snippy.html'
        else:
            pageTemplate = self.viewDir + os.sep + self.pageTemplate
        html = open(pageTemplate, 'r').read()


        strCut1 = '<!--user content starts here-->'
        cut1 = html.find(strCut1)
        strCut2 = '<!--user content ends here-->'
        cut2 = html.find(strCut2, cut1 + len(strCut1)) + len(strCut2)

        self._pageHeader = html[0:cut1]
        self._pageFooter = html[cut2:]


    def doHeader(self, **kw):
        """Sets up a standard page header.

        Default implementation uses HTMl templates. Create an overall
        page template in any web editor and "bracket" it into
        3 sections with comments
             your html header
              <!--user content starts here-->
             any dummy content in the middle
              <!--user content ends here-->
             your html footer

        This should still be viewable in any HTML editor such as
        Dreamweaver.  The TOP portion will become your header, and
        the bottom your footer.

        You may pass in keyword substitutions.  For a given KEY
        and VALUE it will replace
          <!--KEYsnip-->Blah Blah<!--KEYsnip-->
        with
            VALUE.

        """
        if self._pageHeader is None: self._initPageTemplate()
        ppt = kw.get('_preprocessHeader',None)
        if ppt:
            self._pageHeader = ppt(self._pageHeader)
            del kw['_preprocessHeader']
        return _substituteSnip(self._pageHeader, **kw)


    def doFooter(self, **kw):
        """Adds in a standard footer.  See doHeader for default implementation"""
        if self._pageHeader is None: self._initPageTemplate()
        ppt = kw.get('_preprocessFooter',None)
        if ppt:
            self._pageHeader = ppt(self._pageHeader)
            del kw['_preprocessFooter']
        return _substituteSnip(self._pageFooter, **kw)

    ###################################################################
    #
    #   database methods - have no effect unless you need a database
    #
    ###################################################################

    def getConnection(self, name=None):
        "Override this for corporate apps not running in our environment"
        if self._conn is None:
            import reportlab
            if name is None:
                name = self.dbName
                assert name, 'data source name is not specified!'
            self._conn = reportlab.connect(name)
        return self._conn

    def getCursor(self):
        "Returns a DB API cursor"
        if self._cur is None: self._cur = self.getConnection().cursor()
        return self._cur

    def query(self, sql, params=None):
        "Returns data set with minimum fuss"
        self.execute(sql,params)
        return self.getCursor().fetchall()

    def execute(self, sql, params=None):
        "execute an sql statement"
        cur = self.getCursor()
        if params is None:
            cur.execute(sql)
        else:
            assert type(params) in (ListType, TupleType), "Argument params should be a list or tuple"
            cur.execute(sql, params)

    def commit(self):
        self._conn.commit()

    def disconnect(self):
        "Just in case you need explicit disconnection"
        if self._cur is not None:
            self._cur.close()
            self._cur = None
        if self._conn is not None:
            self._conn.close()
            self._conn = None

    ###################################################################
    #
    #   Basic authentication scheme.  Offers a database-aware
    #   implementation.
    #
    ###################################################################
    def authenticate(self, request):
        """This is called on every hit.

        If self.authenticating = 0, it succeeds instantly.
        If self.authenticating = 1, our token mechanism kicks in.
        If you have
           (a) a valid user name and password or
           (b)
        A side effect is that request.user is set after calling
        if there is a user.
        """
        if not self.authenticating:
            return

        success = 0
        userName = None
        failureReason = None
        params = request.params

        # there are 3 main drivers for the logic here, giving 8 states
        hasUserPass = params.has_key('USER') and params.has_key('PASSWORD')
        hasToken = params.has_key('token')
        hasSavedParams = params.has_key('login_saved_params')
        state = hasUserPass + 2* hasToken + 4 * hasSavedParams

        if state == 0:  #nothing
            success = 0
            failureReason = "Login Required"

        elif state == 1: #user and password - new login to default view
            userName = params['USER']
            if self.validateUser(params['USER'], params['PASSWORD']):
                success = 1
            else:
                success = 0
                failureReason = 'Bad username and/or password'

        elif state == 2: #token only
            userName, timedIn = self._check_token(params['token'])
            if timedIn:
                success = 1
            else:
                success = 0
                failureReason = 'Timed out, please log in again'

        elif state == 3: #token, userpass - should not happen
            #trust the token, can the rest
            del params['USER']
            del params['PASSWORD']
            userName, timedIn = self._check_token(params['token'])
            if timedIn:
                success = 1
            else:
                success = 0
                failureReason = 'Timed out, please log in again'

        elif state == 4: #saved params only - should not happen
            success = 0
            failureReason = "Login Required"

        elif state == 5: #user, pass and saved - they tried something
            #else, it saved the parameters and they are logging in.
            #check user.  Also, if saved params has a token,
            #they were here before so check they are not spoofing.
            userName = params['USER']
            if self.validateUser(params['USER'], params['PASSWORD']):
                lspDict = cPickle.loads(base64.decodestring(params['login_saved_params']))
                if lspDict.has_key('token'):
                    #they were logged in before.  Verify the old user was the same.
                    (oldUser, timedIn) = self._check_token(lspDict['token'])
                    if oldUser == userName:
                        success = 1
                    else:
                        success = 0
                        userName = oldUser
                        failureReason = 'You are not %s!' % oldUser
                else:
                    #new login, password OK, let them through
                    success = 1
            else:
                success = 0
                failureReason = 'Bad username and/or password'

        elif state == 6: #token and saved - no idea why but allow it
            userName, timedIn = self._check_token(params['token'])
            if timedIn:
                success = 1
            else:
                success = 0
                failureReason = 'Timed out, please log in again'

        elif state == 7: #everything - ignore password
            del params['USER']
            del params['PASSWORD']
            userName, timedIn = self._check_token(params['token'])
            if timedIn:
                success = 1
            else:
                success = 0
                failureReason = 'Timed out, please log in again'

        # now we know if it worked.  Modify the request
        # accordingly so they go to the right place.

        if success:
            request.user = userName

            #if they were trying to do something before,
            #restore the parameters.  This will typically override the action.
            if params.has_key('login_saved_params'):
                lspDict = cPickle.loads(base64.decodestring(params['login_saved_params']))
                del params['login_saved_params']
                params.update(lspDict)

            if params.has_key('USER'): del params['USER']
            if params.has_key('PASSWORD'): del params['PASSWORD']
            request.params['token'] = self._make_token(userName)
        else:
            loginCount = int(params.get('login_count','0'))
            params['login_count'] = str(loginCount + 1)
            if not hasSavedParams:
                params['login_saved_params'] = base64.encodestring(cPickle.dumps(params,1))[:-1]

            if int(params['login_count']) >= 3:
                params['action'] = 'tooManyLogins'
            else:
                params['action'] = 'showLogin'
                params['USER'] = userName  #may well be None
                if failureReason: params['ERROR_MESSAGE'] = failureReason

        return


    def _get_token(self,request):
        """check for valid user/password and if so return a token.AMPER

        If not valid, return None"""
        try:
            user = request.params['USER']
            password = request.params['PASSWORD']
            del request.params['USER'], request.params['PASSWORD']
        except KeyError:
            return None
        if self.authDummyUsers:
            valid = self._dummyValidateUser(user,password)
        else:
            valid = self.validateUser(user,password)

        if valid:
            request.user  = user
            return self._make_token(user)
        else:
            return None


    def _make_token(self, user, timeout=None):
        from time import time
        from random import random
        timeout = timeout or self.authTimeOut
        token = base64.encodestring(_getRotor().encrypt('%s-%s-%s-%s-%s' % (_get_remote_addr(),time(),random(),timeout,user)))
        return 'aaa-'+token.replace('\n','')

    def _check_token(self,token):
        "Returns (user, timedOutOrNot) pair"
        if token is None:
            return (None, 0)
        if token[:4]!='aaa-': return 0
        token = _getRotor().decrypt(base64.decodestring(token[4:])).split('-')
        if len(token)!=5: return 0
        if _get_remote_addr()!=token[0]: return 0
        try:
            t = float(token[1])
            timeout = float(token[3])
        except:
            return 0
        from time import time
        user = token[4]
        return (user, (time()-t)<=timeout)

    def validateUser(self,user,password):
        """Used by authentication on login.  Return true or
        false for a given username/password pair.  By default
        this uses a database table of md5-encrypted passwords
        in lower case.  Override it to provide alternate schemes.
        This is non-underscored so other code can ask the app
        to validate users.
        """
        if self.authDummyUsers:
            return self._dummyValidateUser(user, password)
        else:
            #standard datyabase scheme
            hashPass = md5.md5(password).hexdigest()
            sql = "select %s from %s where %s='%s' and %s='%s'" % (self.authUserCol,
                            self.authLoginTable,self.authUserCol,user,
                            self.authPasswordCol,hashPass)
            result = self.query(sql)
            if result:
                return 1
            else:
                return 0

    def _dummyValidateUser(self, user, password):
        """Provides 4 canned users so we can test authentication
        without needing a database or files"""
        if user in ['alice','bob','charlie','danny'] and password == user:
            return 1
        else:
            return 0

if __name__=='__main__':
    "Allow a command line test with key=value parameter passing"
    WebApp().handleCommandLine()
