#Copyright ReportLab Europe Ltd. 2000-2004
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/graphics/guiedit/datacharts.py
"""
A uniform GUI-configurable 'data interface' for batch chart creation.

This builds on our concepts of 'data aware drawings'.  The goal is
to let people select CSV, ODBC and ultimately other (e.g. Excel)
data sources for drawings, and generate scripts using the GUI which
can configure what data is associated with what drawing elements.

The directory contains sample CSV files and an Access database
'sampledata.mdb'.  The latter should be registered with the data
source name 'samplechartdata'.  I guess in due course we should
provide a 'canned data' module as a fallback and the ability to
populate any SQL database from it.

We already have 'active drawings': a drawing can have a 'go()'
method, and if so the GUI will generate a script handler from it.

The high speed overview:
We define a 'DataAwareDrawing' which inherits from 'Drawing'.
DataAwareDrawings contain exactly one 'DataSource' widget.
Their 'go' method tells the associated data source to connect,
to repeatedly fetch blocks of data, then close. Each block
of data is a list-of-lists containing data for one drawing.

The data sets  may contain integers, floats, strings, or dates
(all converted to NormalDate objects).  We have no official
policy on filtering out nulls, except to say that ODBC might
return them and it's your job to make sure data is present
where needed. Thus if you want to associate a database column
with a chart title, and some rows would return null, you should
design your stored proc to return an empty string.  If null
is legal in the drawing (e.g. a bar's value can be None
indicating missing data), a null in the data set is OK.

(still to do):
Each data source has a mapping you can specify saying which
columns in the block are associated with which items in the
drawing. Such mappings can be vector or scalar.
Thus you can specify this kind of thing:
 column 1 -> drawing.barchart.title (treated as scalar)
 column 2 -> drawing.barchart.subtitle (ditto)
 column 3 -> drawing.barchart.xAxis.labels
 columns 4,5,6 -> drawing.barchart.data

It should be trivial to extend to allow one to specify
colors in the database

If you don't like the mapping, override the drawing's
applyDataSet method and roll your own.

To make this testable, DataAwareDrawing has a 'test' attribute
set to zero initially.  Change this to (1) and the drawing
will attempt to connect, fetch the data, and apply the first
data block to the drawing.  If it fails, you will get the error
message you deserve (file not found, data association failed
etc.).


"""
__version__=''' $Id: datacharts.py 13212 2004-06-28 12:31:48Z rgbecker $ '''
import string, sys, os

from reportlab.lib import normalDate
from reportlab.lib.attrmap import *
from reportlab.lib.validators import *
from reportlab.lib.validators import _isString, _SequenceTypes  # underscore not imported by default
from reportlab.lib import colors
from reportlab.graphics import shapes
from reportlab.graphics.widgetbase import Widget, TypedPropertyCollection
from reportlab.graphics.charts.piecharts import Pie
from types import ListType, TupleType, StringType, IntType
Drawing = shapes.Drawing

import pprint

#used if you have the database
SAMPLE_DATA_SOURCE = "samplechartdata"
SAMPLE_USER = ''
SAMPLE_PASSWORD = ''

class _isStringOrNone(_isString):
    def test(self,x):
        return _isString.test(self,x) or x is None
isStringOrNone = _isStringOrNone()

class _isValidColumn(Validator):
    "what you can assign to Column"
    def test(self, x):
        if type(x) is IntType:
            return 1
        elif type(x) in _SequenceTypes:
            for E in x:
                if type(E) in _SequenceTypes:
                    for e in E:
                        if type(e) is not  IntType: return 0
                elif type(E) is not IntType: return 0
            return 1
        else:
            return 0

isValidColumn = _isValidColumn()

class Array(Widget):
    """Another kind of collection. Belongs in widgetbase :-)

    Contains exactly n occurrences of its prototype widget,
    all explicitly enumerated as element00, element01 etc."""
    _attrMap = AttrMap(
        size = AttrMapValue(isInt, desc='number of child objects')
        )
    def __init__(self, size, exampleWidgetClass):
        #give it same validation rules as what it holds
        self.__dict__['_prototype'] = exampleWidgetClass
        self.__dict__['_children'] = []
        self.resize(size)

    def __repr__(self):
        return 'Array(%d, %s)' % (self.size, self._prototype.__name__)

    def __setattr__(self, name, value):
        if name == 'size':
            self.resize(value)
        elif name[0:7] == 'element':
            idx = string.atoi(name[7:])
            self._children[idx] = value
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        "This is only called if the attribute is not found"
        if name[0:7] == 'element':
            idx = string.atoi(name[7:])
            return self._children[idx]
        else:
            return Widget.__getattr__(self, name)

    def __getitem__(self, idx):
        return self._children[idx]

    def __setitem__(self, idx, value):
        assert isinstance(value, self._prototype), 'Illegal assignment of %s to array' % value
        self._children[idx] = value

    def __len__(self):
        return len(self._children)

    def draw(self):
        g = shapes.Group()
        for child in self._children:
            g.add(child.draw())
        return g

    def resize(self, newSize):
        # if too big, lose the last few
        while newSize < len(self._children):
            self._children.pop()
        while newSize > len(self._children):
            self._children.append(self._prototype())
        self.__dict__['size'] = newSize


    def getProperties(self, recur=1):
        props = {}
        props['size'] = self.size

        for i in range(self.size):
        #    props['[%d]' % i] = self._children[i]
            props['element%02d' % i] = self._children[i]
        return props

class ArrayOf(Validator):
    def __init__(self,elemTest):
        self._elemTest = elemTest

    def test(self, x):
        if isinstance(x, Array):
            return 0
        else:
            for elem in x._children:
                if not self._elemTest(elem): return 0
            return 1

def rowRavel(pattern, data):
    """Utility for turning one row of data into a nested structure.

    The pattern specifies which indices to use to extract data.append
    List/tuple structures are preserved in the output.  So,
    if the data was the series [0.1, 1.1, 2.1, 3.1 .....19.1],
    rowRavel(2, data) -> 2.1  i.e. return 3rd item
    rowRavel([2,3,4], data) -> [2.1, 3.1, 4.1]
    rowRavel([2,4,6],
             [3,5,7]), data) ->  [[2.1, 4.1, 6.1],
                                 [3.1, 5.1, 7.1]]
    and so on.                
    """
    output = []
    for elem in pattern:
        if type(elem) in _SequenceTypes:
            output.append(rowRavel(elem, data))
        elif type(elem) is IntType:
            output.append(data[elem])
    return output


class DataAssociation(Widget):
    """Defines what a data source column is mapped to in a drawing"""
    _attrMap = AttrMap(
        column = AttrMapValue(isValidColumn, desc='column number (counting from 0) or list of column numbers/sequences'),
        target = AttrMapValue(isStringOrNone, desc='where should the data be applied?'),
        assocType = AttrMapValue(OneOf('scalar','vector','matrix','tmatrix','rowmap'), desc='"scalar", "vector", "matrix" or "tmatrix" or "rowmap"'),
        pyEval = AttrMapValue(isBoolean, desc='If true then use python eval on string values.'),
        )

    def __init__(self, column=0,target=None,assocType='vector',pyEval=0):
        self.column = column
        if type(target) is type(''): target = string.strip(target)
        self.target = target
        if type(assocType) is type(''): assocType = string.strip(assocType)
        self.assocType = assocType
        self.pyEval = pyEval

    def __repr__(self):
        return "%s(column=%s, target=%s, assocType='%s')" % (
                    self.__class__.__name__,
                    self.column,
                    repr(self.target),
                    self.assocType
                    )
    def __len__(self):
        return len(self._children)



    def draw(self):
        """Must be implemented, but returns nothing visual"""
        return shapes.Group()

class _isDataAssociation(Validator):
    "Is it a descendant of DataAssociation?"
    def test(self, x):
        return isinstance(x, DataAssociation)
isDataAssociation = _isDataAssociation()
isArrayOfAssociations = ArrayOf(isDataAssociation)

from rlextra.utils.datamaker import DataMaker, CSVDataMaker
class DataSource(DataMaker,Widget):
    """Editable component to define data acquisition"""
    # clunky hack for now - allow up to 10 columns
    _attrMap = AttrMap(
        associations = AttrMapValue(isAnything, desc='drawing elements for each data ')
        )

    def __init__(self):
        DataMaker.__init__(self)
        self.associations = Array(0, DataAssociation)

    def draw(self):
        """Must be implemented, but returns nothing visual"""
        return shapes.Group()

class _isDataSource(Validator):
    "Is it a descendant of DataSource?"
    def test(self, x):
        return isinstance(x, DataSource)
isDataSource = _isDataSource()

class DummyPieDataSource(DataSource):
    """Some fake data for a trivial test"""
    def connect(self):
        self._rows = [(1,2),
                     (1,2,3),
                     (1,2,3,4,)]

    def fetchNextDataSet(self):
        if self._rows:
            head = self._rows[0]
            self._rows = self._rows[1:]
            return head
        else:
            return None

class ODBCDataSource(DataSource):
    """Generic data source for ODBC connections"""
    _attrMap = AttrMap(
            BASE=DataSource,
            name = AttrMapValue(isString, desc='ODBC Data Source Name'),
            user = AttrMapValue(isString, desc='User Name (None if not required)'),
            password = AttrMapValue(isString, desc='Password (None if not required)'),
            sql = AttrMapValue(isStringOrCallable, desc='SQL Statement or Stored Procedure'),
            groupingColumn = AttrMapValue(isInt, desc='Column to group charts on (starts from 0)')
            )

    def __init__(self):
        DataSource.__init__(self)
        self.name = SAMPLE_DATA_SOURCE
        self.user = SAMPLE_USER
        self.password = SAMPLE_PASSWORD
        self.sql = 'SELECT * FROM generic_pie'
        self.groupingColumn = 0


    def connect(self, verbose=0):
        import mx.ODBC.Windows
        conn = mx.ODBC.Windows.connect(self.name, self.user, self.password,
                                       clear_auto_commit=0)
        try:
            cur = conn.cursor()
            sql = self.sql
            if callable(sql): sql = sql()
            cur.execute(sql)
            self._data = cur.fetchall()
            if verbose:
                print 'fetched %d rows' % len(self._data)
            
            self.convertDates()
            self._description = cur.description[:]  # might be useful
            cur.close()
            conn.close()

            # diagnostic hack - first row of data
##          print 'dumping table structure...'
##          colCount = len(self._description)
##          print "colCount: %s" % colCount
##          if len(self._data) == 0:
##              print 'no data returned'
##          else:
##              print "len(self._data): %s" % len(self._data)
##          for col in range(colCount):
##              descr = self._description[col][0]
##              try:
##                  value = self._data[0][col]
##                  valType = type(value)
##                  print 'col %d: %s = %s (type %s)' % (col, descr, value, valType)
##              except: # probably no data
##                  print 'col %d: %s (no data available)' % (col, descr)
                    
        except:
            #make sure connections closed
            conn.close()
            raise

    def convertDates(self):
        # convert all mx.DateTime objects to normalDate objects
        import mx.DateTime
        nd = normalDate.NormalDate
        newdata = []
        dateTyp = type(mx.DateTime.now())
        for row in self._data:
            newRow = []
            for field in row:
                if type(field) is dateTyp:
                    newRow.append(nd((field.year, field.month, field.day)))
                else:
                    newRow.append(field)
            newdata.append(tuple(newRow))
        self._data = newdata

    def close(self):
        # not needed as we fetched the lot anyway
        pass

isColumn=EitherOr((isNumber,isString),name='isColumn')
isNoneOrColumnList=SequenceOf(isColumn,'isNoneOrColumnList',NoneOK=1)
class CSVDataSource(CSVDataMaker,DataSource):
    """Defines how to parse a CSV file"""
    _attrMap = AttrMap(
            BASE=DataSource,
            filename = AttrMapValue(isString, desc='filename'),
            headerRows = AttrMapValue(isInt, desc='Number of header rows at start of file?'),
            nameRow = AttrMapValue(isNoneOrInt, desc='Which header row(zero based) to use for column names (default None)?'),
            columnCount = AttrMapValue(isInt, desc='How many columns the file contains'),
            groupingColumn = AttrMapValue(isColumn, desc='Column to group charts on (starts from 0)'),
            integerColumns = AttrMapValue(isNoneOrColumnList, desc='Zero based columns containing integers'),
            floatColumns = AttrMapValue(isNoneOrColumnList, desc='Zero based columns containing floating point numbers'),
            dateColumns = AttrMapValue(isNoneOrColumnList, desc='Zero based columns containing dates'),
            sep = AttrMapValue(isString, desc='The separator value, default ","'),
            fast = AttrMapValue(isBoolean, desc='If true use the fast not safe split (default off)'),
            sql = AttrMapValue(isStringOrCallableOrNone, desc='SQL Statement'),
            na = AttrMapValue(isListOfStringsOrNone, desc='Case insensitive strings which map to None'),
            sbl = AttrMapValue(isBoolean, desc='If true skip blank lines'),
            )

    def __init__(self):
        DataSource.__init__(self)
        CSVDataMaker.__init__(  self,
                                filename = 'lineplots.csv',
                                #headerRows = 1,
                                #nameRow = 0,
                                #columnCount = 0,
                                #groupingColumn = 0,
                                #integerColumns = (),
                                #dateColumns = (),
                                #floatColumns = (),
                                #sep = ',',
                                #fast = 0,
                                #na = ['None', 'N/A'],
                                #sbl = 1,
                                #sql = None,
                                )

def _convert2CSV(D,csvFN):
    '''Convert an ODBC to CSV'''
    A = D.dataSource.associations
    sql = D.dataSource.sql
    DD = D.dataSource = CSVDataSource()
    DD.associations = A
    DD.filename = csvFN
    DD.nameRow = 0
    DD.headerRows = 1
    DD.sql = sql
    colInfo = getattr(D,'_colInfo',{})
    for x in ('float', 'integer', 'date'):
        if colInfo.has_key(x):
            v = colInfo[x]
        else:
            v = []
        setattr(DD,x+'Columns',v)

def _pyEval(v):
    t = type(v)
    if t is StringType:
        try:
            return eval('"""%s"""'%v,locals())
        except:
            return v
    elif t in [ListType,TupleType]:
        v = map(_pyEval,v)
        if t is TupleType: v = tuple(v)
        return v
    else:
        return v

class DataAwareDrawing(Drawing):
    """Mixin for adding data capabilities and main loop to a drawing"""
    _attrMap = AttrMap(
            BASE=Drawing,
            dataSource = AttrMapValue(isDataSource, desc='Where to get the data'),
            test = AttrMapValue(isInt, desc='set to 1 to test connection'),
            sampleData = AttrMapValue(isAnything, desc='whatever gets fetched...'),
            outDir = AttrMapValue(isString, desc='path where files should be written (default .)'),
            fileNamePattern = AttrMapValue(isString, desc='pattern for naming output files'),
            chartId = AttrMapValue(isAnything, desc='identifier for this chart'),
            EPS_info = AttrMapValue(SequenceOf(isString,lo=2,hi=2,emptyOK=0), desc='Dept and Company for EPS header info'),
            preview = AttrMapValue(isNumber, desc='Non zero for preview>1 for scaling'),
            formats = AttrMapValue(SequenceOf(OneOf('eps','pdf','png','gif','jpg','tif'),lo=1,emptyOK=0), desc='One or more plot modes'),
            showBorder=AttrMapValue(isBoolean, desc='If true showBorder when possible'),
            verbose = AttrMapValue(isBoolean, desc='Boolean, true for verbose operation'),
            )

    def __init__(self, width=400, height=200,*nodes,**kw):
        Drawing.__init__(self,width,height,*nodes,**kw)
        self.dataSource = DataSource()
        self.test = 0
        self.sampleData = None
        self.outDir = os.curdir
        self.fileNamePattern = 'drawing%03d'
        self.chartId = 0
        self.EPS_info = ('SFR','Fidelity')
        self.preview = 1
        self.formats = ['eps','gif']
        self.showBorder = 0
        self.verbose = 0

    def print_versions(self):
        if self.quiet:
            return
        def CVSVer(s):
            s=string.strip(string.replace(s, ' Exp ', ''))
            return string.strip(string.join(string.split(s)[1:-1]))
        import reportlab
        print "do_graph  Version %s, %s" % (Version, CVSVer(__version__))
        if self.verbose:
            print "reportlab Version %s, %s"% (reportlab.Version, CVSVer(reportlab.__version__))

    def Usage(self):
        F = os.path.split(os.path.abspath(sys.argv[0]))
        print 'Usage: %s [options] [configs....]' % F[1]
        print ' options, -q quiet, -v verbose, -V version only,'
        print ' -mode=[eps,pdf,tiff,gif,png,jpg]'
        print ' -scale=number'
        print ' -h this help\n'
        print ' configs are files like do_graphs.ini'
        print ' if none are given then do_graphs.ini in'
        print ' directory "%s" is used.' % F[0]

    def __setattr__(self, name, value):
        "Hook to allow testing in the GUI"
        if name == 'test' and value >= 1:
            self.testConnect(value)
        self.__dict__[name] = value

    def testConnect(self,value):
        "Called when test set to 1 e.g. in the GUI"
        src = self.dataSource
        src.connect()
        while value>0:
            sample = src.fetchNextDataSet()
            value = value - 1
        self.__dict__['sampleData'] = sample
        src.close()
        self.applyDataSet(sample)

    def applyDataSet(self, dataSet):
        """By default, uses associations to extract and match data.

        If you have not defined any, nothing happens.  For now each
        association is either a tuple (type, target) or just a target
        expression in the drawing.  """
        #print 'applying Data Set:'
        #pprint.pprint(dataSet)

                
        for assoc in self.dataSource.associations:
            if assoc.target is not None:
                #determine the value - might be a list, might be a scalar
                value = None
                if assoc.assocType == 'scalar':
                    #  row 1
                    if len(dataSet) == 0:
                        value = None
                    else:
                        value = dataSet[0][assoc.column]
                elif assoc.assocType == 'vector':
                    value = []
                    C = assoc.column
                    if type(C) in _SequenceTypes:
                        n = 0
                        for row in dataSet:
                            R = []
                            for c in C:
                                R.append(row[c])
                            if filter(lambda x: x is not None,R)!=[]: n = n + 1
                            value.append(tuple(R))
                        if n==0: value = None
                    else:
                        for row in dataSet:
                            value.append(row[C])
                        if filter(lambda x: x is not None,value)==[]: value = None
                elif assoc.assocType == 'matrix':
                    value = []
                    C = assoc.column
                    if type(C) not in _SequenceTypes: C = [C]
                    for row in dataSet:
                        selected = []
                        nonNone = 0
                        for G in C:
                            if type(G) in _SequenceTypes:
                                e = []
                                n = 0
                                for c in G:
                                    y = row[c]
                                    e.append(y)
                                    n += y!=None
                                selected.append(tuple(e))
                                nonNone += n==len(G)
                            else:
                                y = row[G]
                                selected.append(y)
                                nonNone += y is not None
                        if nonNone:
                            value.append(tuple(selected))
                    value = tuple(value)
                elif assoc.assocType == 'tmatrix':
                    value = []
                    C = assoc.column
                    if type(C) not in _SequenceTypes: C = [C]
                    for G in C:
                        selected = []
                        nonNone = 0
                        if type(G) in _SequenceTypes:
                            for row in dataSet:
                                e = []
                                n = 0
                                for c in G:
                                    y = row[c]
                                    e.append(y)
                                    n += y!=None
                                selected.append(tuple(e))
                                nonNone += n==len(G)
                        else:
                            for row in dataSet:
                                y = row[G]
                                selected.append(y)
                                nonNone += y is not None
                        if nonNone:
                            value.append(tuple(selected))
                    value = tuple(value)
                elif assoc.assocType == 'rowmap':
                    # for one-row data sets - a perversion I hope we encounter only
                    # rarely.
                    row = dataSet[0]
                    pat = assoc.column
                    value = rowRavel(pat, row)

                if assoc.pyEval: value = _pyEval(value)
                # traverse to the target
                path = map(string.strip,string.split(assoc.target, '.'))
                if len(path) == 0:
                    #empty string
                    return
                elif len(path) == 1:
                    # they are applying to top level drawing
                    setattr(self, path[0], value)
                else:
                    #trace through object hierarchy
                    cursor = getattr(self, path[0])
                    for elem in path[1:-1]:
                        cursor = getattr(cursor, elem)
                    setattr(cursor, path[-1], value)

    def _saveHook(self):
        self.save()

    def go(self, verbose=0, debug=None):
        try:
            if verbose:
                print 'connecting...',
            self.dataSource.connect(verbose=verbose)
            if verbose:
                print 'OK'
            r = 0
            while 1:
                fetched = None
                r = r + 1
                fetched = self.dataSource.fetchNextDataSet()
                if not fetched:
                    break
                self.applyDataSet(fetched)
                ##print 'self.chart.data=',
                ##pprint.pprint(self.chart.data)
                #print 'self.chartId = %s' % self.chartId
                self._saveHook()
            self.dataSource.close()
        except:
            if debug:
                from reportlab.lib.utils import DebugMemo
                dbg = DebugMemo(
                        fn=os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])),
                                    'RL_'+self.__class__.__name__+'_pickle.dbg'),
                        class_name = self.__class__.__name__,
                        fetched = fetched,
                        recordNumber = r,
                        _data=getattr(self,'_data',None)
                        )
                dbg.dump()
                print >>sys.stderr, 'A DebugMemo has been written to "%s"' % dbg.fn
            raise

class DataAwarePieDrawing(DataAwareDrawing):
    """Trivial example of a data drawing.  """
    def __init__(self,width=200,height=150,*nodes,**kw):
        hasDB = 0
        if kw.get('hasDB',0) == 1:
            hasDB = 1

        DataAwareDrawing.__init__(self,width,height,*nodes,**kw)
        p = Pie()
        p.x = 50
        p.y = 10
        self.add(p, 'pie')

        # add a title to test scalar associations
        self.add(shapes.String(100,135,'Chart Title',
                        fontName='Times-Bold',
                        fontSize=14,
                        textAnchor='middle'),
                        'title'
                        )
        self.outDir = 'output'
        self.fileNamePattern = 'datapie%03d'
        self.formats = ['eps','pdf','gif']

        #in real life we'd set this up in the GUI
        if hasDB:
            self.dataSource = ODBCDataSource()
            self.dataSource.name = 'samplechartdata'
            self.dataSource.sql = 'SELECT ChartId, SliceId, Name, Value, Title FROM generic_pie'
        else:
            self.dataSource = CSVDataSource()
            self.dataSource.filename = 'samplepie.csv'
            self.dataSource.headerRow = 1
            self.dataSource.columnCount = 5
            self.dataSource.groupingColumn = 0
            self.dataSource.integerColumns = (1,2)
            self.dataSource.dateColumns = ()
            self.dataSource.floatColumns = (4,)

        self.dataSource.associations = Array(4, DataAssociation)

        self.dataSource.associations.element00.column = 0
        self.dataSource.associations.element00.target = 'chartId'
        self.dataSource.associations.element00.assocType = 'scalar'


        self.dataSource.associations.element01.column = 2
        self.dataSource.associations.element01.target = 'pie.labels'

        self.dataSource.associations.element02.column = 3
        self.dataSource.associations.element02.target = 'pie.data'

        self.dataSource.associations.element03.column = 4
        self.dataSource.associations.element03.target = 'title.text'
        self.dataSource.associations.element03.assocType = 'scalar'

        self.verbose = 1

def test():
    if not os.path.isdir('output'):
        os.mkdir('output')
    # can we use the database?
    import sys
    if '-db' in sys.argv:
        db = 1
    elif '-csv' in sys.argv:
        db = 0
    else:
        try:
            import mx.ODBC.Windows
            conn = mx.ODBC.Windows.connect('samplechartdata')
            conn.close()
            db = 1
        except:
            db = 0

    #make a batch of drawings and write out
    if db:
        print 'using database as data source...'
    else:
        print 'using CSV file as data source...'

    d = DataAwarePieDrawing(hasDB=db)
    d.go()
    class DINGO(DataAwareDrawing):
        _attrMap = AttrMap(BASE=DataAwareDrawing,
            data = AttrMapValue(None, desc='dummy data value'),
            )
        def go(self):
            print self.dataSource.sql
            A = self.dataSource.associations
            for a in A:
                print 'column=%s, target=%s, assocType=%s' % (str(a.column),repr(a.target), repr(a.assocType))
            self.dataSource.connect()
            while 1:
                fetched = self.dataSource.fetchNextDataSet()
                if not fetched:
                    break
                self.applyDataSet(fetched)
                for a in A:
                    print '%s:' % a.target,
                    pprint.pprint(getattr(self,a.target))
                print
                self.save()
            self.dataSource.close()

    d = DINGO()
    d.dataSource = ODBCDataSource()
    d.dataSource.sql  = 'SELECT Id, C1, C2, C3, C4 FROM example'
    d.dataSource.associations   = Array(2, DataAssociation)
    d.dataSource.associations.element00 = DataAssociation(column=0, target='chartId', assocType='scalar')
    d.dataSource.associations.element01 = DataAssociation(column=[[1, 2], [1,3], [1,4]], target='data', assocType='tmatrix')
    d.formats = []
    d.go()
    d = DINGO()
    d.dataSource = ODBCDataSource()
    d.dataSource.sql  = 'SELECT Id, C1, C2, C3, C4 FROM example'
    d.dataSource.associations   = Array(4, DataAssociation)
    d.dataSource.associations.element00 = DataAssociation(column=0, target='chartId', assocType='scalar')
    d.dataSource.associations.element01 = DataAssociation(column=1, target='cat', assocType='vector')
    d.dataSource.associations.element02 = DataAssociation(column=[2], target='data', assocType='tmatrix')
    d.dataSource.associations.element03 = DataAssociation(column=4, target='label', assocType='vector')
    d.go()
    d = DINGO()
    d.dataSource = ODBCDataSource()
    d.dataSource.sql  = 'SELECT Id, C1, C2 FROM example'
    d.dataSource.associations   = Array(2, DataAssociation)
    d.dataSource.associations.element00 = DataAssociation(column=0, target='chartId', assocType='scalar')
    d.dataSource.associations.element01 = DataAssociation(column=1, target='data', assocType='vector')
    d.go()
    d = DINGO()
    d.dataSource = ODBCDataSource()
    d.dataSource.sql  = 'SELECT Id, C1, C2 FROM example'
    d.dataSource.associations   = Array(2, DataAssociation)
    d.dataSource.associations.element00 = DataAssociation(column=0, target='chartId', assocType='scalar')
    d.dataSource.associations.element01 = DataAssociation(column=[1,2], target='data', assocType='vector')
    d.go()


if __name__=='__main__':
    test()
